from sl_ai_models.basic_model_interfaces.retryable_model import RetryableModel
from sl_ai_models.basic_model_interfaces.tokens_incur_cost import TokensIncurCost
from sl_ai_models.basic_model_interfaces.time_limited_model import TimeLimitedModel
from sl_ai_models.basic_model_interfaces.request_limited_model import RequestLimitedModel
from sl_ai_models.basic_model_interfaces.token_limited_model import TokenLimitedModel
from sl_ai_models.basic_model_interfaces.outputs_text import OutputsText
from sl_ai_models.basic_model_interfaces.named_model import NamedModel

from abc import ABC
from typing import Any
import logging
logger = logging.getLogger(__name__)


class TraditionalOnlineLlm(
    TokenLimitedModel,
    RequestLimitedModel,
    TimeLimitedModel,
    TokensIncurCost,
    RetryableModel,
    OutputsText,
    NamedModel,
    ABC
    ):


    def __init__(
            self,
            temperature: float = 0,
            allowed_tries: int = RetryableModel._DEFAULT_ALLOWED_TRIES,
            system_prompt: str | None = None,
            api_key: str | None = None
        ) -> None:
        super().__init__(allowed_tries=allowed_tries)
        assert temperature >= 0, "Temperature must be greater than or equal to 0"
        assert temperature <= 1, "Temperature must be less than or equal to 1"
        self.temperature: float = temperature
        self.system_prompt: str | None = system_prompt
        self.api_key: str | None = api_key


    async def invoke(self, *args, **kwargs) -> Any:
        result = await self._invoke_with_request_cost_time_and_token_limits_and_retry(*args, **kwargs)
        return result


    @RequestLimitedModel._wait_till_request_capacity_available
    @TokenLimitedModel._wait_till_token_capacity_available
    @RetryableModel._retry_according_to_model_allowed_tries
    @TokensIncurCost._wrap_in_cost_limiting_and_tracking
    @TimeLimitedModel._wrap_in_model_defined_timeout
    async def _invoke_with_request_cost_time_and_token_limits_and_retry(self, *args, **kwargs) -> Any:
        logger.debug(f"Invoking model with args: {args} and kwargs: {kwargs}")
        direct_call_response = await self._mockable_direct_call_to_model(*args, **kwargs)
        response_to_log = direct_call_response[:1000] if isinstance(direct_call_response, str) else direct_call_response
        logger.debug(f"Model responded with: {response_to_log}...")
        return direct_call_response



    @classmethod
    def _initialize_rate_limiters(cls) -> None:
        cls._reinitialize_request_rate_limiter()
        cls._reinitialize_token_limiter()





import asyncio
from typing import Any, Coroutine
import nest_asyncio
import logging
from aiolimiter import AsyncLimiter
import asyncio
from typing import Coroutine, Any, Callable, TypeVar
logger = logging.getLogger(__name__)

T = TypeVar('T')
T2 = TypeVar('T2')

nest_asyncio.apply() # Make sure asyncio can be nested inside of other asyncio calls


def wrap_coroutines_with_rate_limit(coroutine_list: list[Coroutine[Any,Any,T]], calls_per_period: int, time_period_in_seconds: int = 60) -> list[Coroutine[Any,Any,T]]:
    """
    Rate Limiting is only applied to the coroutines in the list, and not between calls of this function.
    """
    limiter = AsyncLimiter(max_rate=calls_per_period, time_period=time_period_in_seconds)
    return [apply_limiter_to_coroutine(coroutine, limiter) for coroutine in coroutine_list]


async def apply_limiter_to_coroutine(coroutine: Coroutine, limiter: AsyncLimiter) -> Any:
    await limiter.acquire()
    return await coroutine


def wrap_coroutines_with_timeout(coroutine_list: list[Coroutine[Any,Any,T]], timeout_time: float) -> list[Coroutine[Any,Any,T]]:
    async def coroutine_with_timeout(coroutine: Coroutine, timeout_time: float) -> Any:
        try:
            result = await asyncio.wait_for(coroutine, timeout=timeout_time)
            return result
        except asyncio.TimeoutError as e:
            raise asyncio.TimeoutError(f"Timeout of {timeout_time} seconds exceeded while running coroutine. Here is the exception: {e.__class__.__name__}: {e}")
        except Exception as e:
            raise RuntimeError(f"Exception while running coroutine with timeout wrapper. Here is the exception: {e.__class__.__name__}: {e}")
    return [coroutine_with_timeout(coroutine, timeout_time) for coroutine in coroutine_list]


def wrap_coroutines_to_return_not_raise_exceptions(coroutine_list: list[Coroutine[Any,Any,T]]) -> list[Coroutine[Any,Any, T | Exception]]:
    async def coroutine_where_exception_is_returned_not_raised(coroutine: Coroutine) -> Any | Exception:
        try:
            return await coroutine
        except Exception as e:
            return e
    return [coroutine_where_exception_is_returned_not_raised(coroutine) for coroutine in coroutine_list]


def wrap_coroutines_with_limit_timeout_and_returning_exceptions(coroutine_list: list[Coroutine[Any,Any,T]], calls_per_period: int, time_period: int = 60, timeout_time: float = 120) -> list[Coroutine[Any,Any, T | Exception]]:
    rate_limited_coroutines = wrap_coroutines_with_rate_limit(coroutine_list, calls_per_period, time_period)
    limited_and_timed_coroutines = wrap_coroutines_with_timeout(rate_limited_coroutines, timeout_time)
    limted_timed_error_handled_coroutines = wrap_coroutines_to_return_not_raise_exceptions(limited_and_timed_coroutines)
    return limted_timed_error_handled_coroutines


def run_coroutines(coroutines: list[Coroutine[Any,Any,T]]) -> list[T]:
    async def run_coroutines(coroutines: list[Coroutine[Any,Any,T]]) -> list[T]:
        tasks = []
        for coroutine in coroutines:
            tasks.append(loop.create_task(coroutine))
        results = await asyncio.gather(*tasks)
        return results

    loop = asyncio.get_event_loop()
    return loop.run_until_complete(run_coroutines(coroutines))


def run_coroutines_while_removing_and_logging_exceptions(
        coroutines: list[Coroutine[Any,Any,T]],
        matching_inputs: list[T2] | T2 = None,
        action_on_exception: Callable[[Exception, T2], None] | None = None
    ) -> tuple[list[T], list[T2]]:
    """
    Runs a list of coroutines and returns only the results (and their corresponding inputs) that did not raise an exception.
    A list of "None" is returned as the corresponding input if no matching_inputs are provided.
    An default log message is given on the case of an exception. You can switch out this with a cutom function if desired.
    """

    if matching_inputs is None:
        modified_inputs = [None] * len(coroutines)
    elif not isinstance(matching_inputs, list):
        modified_inputs = [matching_inputs] * len(coroutines)
    else:
        modified_inputs = matching_inputs

    assert len(modified_inputs) == len(coroutines), "The number of inputs must match the number of coroutines"

    exception_wrapped_coroutines = wrap_coroutines_to_return_not_raise_exceptions(coroutines)
    results = run_coroutines(exception_wrapped_coroutines)

    results_that_did_not_error: list[T] = []
    inputs_that_did_not_error: list[T2] = []
    for input, result, coroutine in zip(modified_inputs, results, coroutines):
        if isinstance(result, Exception):
            error = result
            if action_on_exception is None:
                action_on_exception = lambda error, _, coroutine = coroutine: logger.error(f"Error while running coroutine '{coroutine.cr_code.co_name}': {error.__class__.__name__} Exception - {error}")
            action_on_exception(error, input) # type: ignore - Linter improperly thinks that input can't be of type 'None' even if None is assigned to Generic type. It works if the default value for inputs is set to an int
        else:
            results_that_did_not_error.append(result)
            inputs_that_did_not_error.append(input) # type: ignore - Linter improperly thinks that input can't be of type 'None' even if None is assigned to Generic type. It works if the default value for inputs is set to an int

    return results_that_did_not_error, inputs_that_did_not_error



from sl_ai_models.utils.response_types import TextTokenCostResponse
from sl_ai_models.model_archetypes.traditional_online_llm import TraditionalOnlineLlm

from langchain_community.callbacks.bedrock_anthropic_callback import _get_anthropic_claude_token_cost, MODEL_COST_PER_1K_INPUT_TOKENS
from langchain_anthropic import ChatAnthropic
from langchain_core.messages import HumanMessage, SystemMessage, BaseMessage

from abc import ABC
import logging

logger = logging.getLogger(__name__)


class AnthropicTextToTextModel(TraditionalOnlineLlm, ABC):

    def __init__(self, max_tokens: int = 1024, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        assert max_tokens > 0, "Max tokens must be greater than 0"
        self.max_tokens = max_tokens


    async def invoke(self, prompt: str) -> str:
        response: TextTokenCostResponse =  await self._invoke_with_request_cost_time_and_token_limits_and_retry(prompt)
        return response.data


    async def _mockable_direct_call_to_model(self, prompt: str) -> TextTokenCostResponse:
        self._everything_special_to_call_before_direct_call()
        response: TextTokenCostResponse = await self._call_online_model_using_api(prompt)
        return response


    async def _call_online_model_using_api(self, prompt: str) -> TextTokenCostResponse:
        anthropic_llm = ChatAnthropic(
            model_name=self.MODEL_NAME,
            temperature=self.temperature,
            api_key=self.api_key,
            timeout=None,
            stop=None,
            base_url=None,
            max_tokens=self.max_tokens, #type: ignore
        )
        messages = self._turn_model_input_into_messages(prompt)
        answer_message = await anthropic_llm.ainvoke(messages)
        answer = answer_message.content

        response_metatdata = answer_message.response_metadata
        prompt_tokens = response_metatdata["usage"]["input_tokens"]
        completion_tokens = response_metatdata["usage"]["output_tokens"]
        total_tokens = prompt_tokens + completion_tokens
        cost = self.caculate_cost_from_tokens(prompt_tkns=prompt_tokens, completion_tkns=completion_tokens)

        assert isinstance(answer, str), "Answer is not a string"
        assert cost >= 0, "Cost is less than 0"
        assert prompt_tokens >= 0, "Prompt Tokens is less than 0"
        assert completion_tokens >= 0, "Completion Tokens is less than 0"

        return TextTokenCostResponse(
            data=answer,
            prompt_tokens_used=prompt_tokens,
            completion_tokens_used=completion_tokens,
            total_tokens_used=total_tokens,
            model=self.MODEL_NAME,
            cost=cost
        )


    def _turn_model_input_into_messages(self, prompt: str) -> list[BaseMessage]:
        if self.system_prompt is None:
            return [HumanMessage(prompt)]
        else:
            return [
                SystemMessage(self.system_prompt),
                HumanMessage(prompt)
            ]


    ################################## Methods For Mocking/Testing ##################################


    @classmethod
    def _get_mock_return_for_direct_call_to_model_using_cheap_input(cls) -> TextTokenCostResponse:
        cheap_input = cls._get_cheap_input_for_invoke()
        probable_output = "Hello! How can I assist you today? Feel free to ask any questions or let me know if you need help with anything."
        if "opus" in cls.MODEL_NAME.lower():
            probable_output = "Hello! How can I assist you today?"

        model = cls()
        prompt_tokens = model.input_to_tokens(cheap_input)
        anthropic_llm = ChatAnthropic(
            model_name=model.MODEL_NAME,
            api_key=None,
            timeout=None,
            stop=None,
            base_url=None,
        )
        completion_tokens = anthropic_llm.get_num_tokens(probable_output)
        adjustment = 3 # Through manual experimentation, it was found that the number of tokens returned by the API is 3 off for the completion response
        completion_tokens += adjustment
        total_cost = model.caculate_cost_from_tokens(prompt_tkns=prompt_tokens, completion_tkns=completion_tokens)
        total_tokens = prompt_tokens + completion_tokens
        return TextTokenCostResponse (
            data=probable_output,
            prompt_tokens_used=prompt_tokens,
            completion_tokens_used=completion_tokens,
            total_tokens_used=total_tokens,
            model=cls.MODEL_NAME,
            cost=total_cost
        )


    @staticmethod
    def _get_cheap_input_for_invoke() -> str:
        return "Hi"



    ############################# Cost and Token Tracking Methods #############################

    def input_to_tokens(self, prompt: str) -> int:
        langchain_llm = ChatAnthropic(
            model_name=self.MODEL_NAME,
            api_key=None,
            timeout=None,
            stop=None,
            base_url=None
        )
        messages = self._turn_model_input_into_messages(prompt)
        tokens = langchain_llm.get_num_tokens_from_messages(messages)
        adjustment = 0
        for message in messages:
            if isinstance(message, HumanMessage):
                adjustment += 5  # Through manual experimentation, it was found that the number of tokens returned by the API is 5 off for user messages
            if isinstance(message, SystemMessage):
                adjustment -= 2  # Through manual experimentation, it was found that the number of tokens returned by the API is 2 off for system messages
        tokens += adjustment
        return tokens


    def caculate_cost_from_tokens(self, prompt_tkns: int, completion_tkns: int) -> float:
        # possible_detailed_model_names = MODEL_COST_PER_1K_INPUT_TOKENS.keys()
        # detailed_model_name = [name for name in possible_detailed_model_names if self.MODEL_NAME in name][0]
        # cost = _get_anthropic_claude_token_cost(prompt_tkns, completion_tkns, detailed_model_name)
        # return cost

        # NOTE: caculate_cost_from_tokens is not implemented since langchain is failing to track the model's cost for models like opus
        return 0






from abc import ABC, abstractmethod
from typing import Any, Callable, Coroutine, TypeVar
import functools
from sl_ai_models.resource_managers.monetary_cost_manager import MonetaryCostManager
T = TypeVar('T')


class IncursCost(ABC):
    """
    Interface indicating a model should track its cost in the monetary cost manager.
    """

    @abstractmethod
    def _estimate_before_plus_after_cost_from_model_input(self, *args, **kwargs) -> float:
        pass


    @abstractmethod
    async def _track_cost_in_manager_using_model_response(self, response_from_direct_call: Any, original_cost_prediction: Any) -> None:
        pass


    @staticmethod
    def _wrap_in_cost_limiting_and_tracking(func: Callable[..., Coroutine[Any, Any, T]]) -> Callable[..., Coroutine[Any, Any, T]]:
        @functools.wraps(func)
        async def wrapper(self: IncursCost, *args, **kwargs) -> T:
            predicted_cost = self._estimate_before_plus_after_cost_from_model_input(*args, **kwargs)
            prediction_reciept = MonetaryCostManager.add_usage_prediction_to_parent_managers(predicted_cost)
            MonetaryCostManager.raise_error_if_limit_would_be_reached()

            direct_call_response = await func(self, *args, **kwargs)

            await self._track_cost_in_manager_using_model_response(direct_call_response, prediction_reciept)
            return direct_call_response
        return wrapper





from sl_ai_models.gpt4ovision import Gpt4oVision
from sl_ai_models.gpt4o import Gpt4o
from sl_ai_models.perplexity import Perplexity
from sl_ai_models.claude35sonnet import Claude35Sonnet
from sl_ai_models.exa_searcher import ExaSearcher
from sl_ai_models.gpt4omini import Gpt4oMini
from sl_ai_models.deepgram_nova2 import DeepgramNova2
from sl_ai_models.claude3opus import Claude3Opus
from sl_ai_models.model_archetypes.anthropic_text_model import AnthropicTextToTextModel

from sl_ai_models.basic_model_interfaces.token_limited_model import TokenLimitedModel
from sl_ai_models.basic_model_interfaces.retryable_model import RetryableModel
from sl_ai_models.basic_model_interfaces.time_limited_model import TimeLimitedModel
from sl_ai_models.basic_model_interfaces.request_limited_model import RequestLimitedModel
from sl_ai_models.basic_model_interfaces.incurs_cost import IncursCost
from sl_ai_models.basic_model_interfaces.ai_model import AiModel
from sl_ai_models.basic_model_interfaces.outputs_text import OutputsText
from sl_ai_models.basic_model_interfaces.tokens_incur_cost import TokensIncurCost


class ModelsToTest:
    ALL_MODELS = [
        Gpt4o,
        Gpt4oMini,
        Gpt4oVision,
        Perplexity,
        Claude35Sonnet,
        Claude3Opus,
        ExaSearcher,
        DeepgramNova2,
    ]
    BASIC_MODEL_LIST: list[type[AiModel]] = [model for model in ALL_MODELS if issubclass(model, AiModel)]
    RETRYABLE_LIST: list[type[RetryableModel]] = [model for model in ALL_MODELS if issubclass(model, RetryableModel)]
    TIME_LIMITED_LIST: list[type[TimeLimitedModel]] = [model for model in ALL_MODELS if issubclass(model, TimeLimitedModel)]
    REQUEST_LIMITED_LIST: list[type[RequestLimitedModel]] = [model for model in ALL_MODELS if issubclass(model, RequestLimitedModel)]
    TOKEN_LIMITED_LIST: list[type[TokenLimitedModel]] = [model for model in ALL_MODELS if issubclass(model, TokenLimitedModel)]
    OUTPUTS_TEXT: list[type[OutputsText]] = [model for model in ALL_MODELS if issubclass(model, OutputsText)]

    INCURS_COST_LIST: list[type[IncursCost]] = [
        model
        for model in ALL_MODELS
        if issubclass(model, IncursCost) and not issubclass(model, AnthropicTextToTextModel)
    ]
    TOKENS_INCUR_COST_LIST: list[type[TokensIncurCost]] = [
        model
        for model in ALL_MODELS
        if issubclass(model, TokensIncurCost)
    ]




from __future__ import annotations

from sl_ai_models.utils.jsonable import Jsonable
from sl_ai_models.basic_model_interfaces.request_limited_model import RequestLimitedModel
from sl_ai_models.basic_model_interfaces.retryable_model import RetryableModel
from sl_ai_models.basic_model_interfaces.time_limited_model import TimeLimitedModel
from sl_ai_models.basic_model_interfaces.priced_per_request import PricedPerRequest
from sl_ai_models.basic_model_interfaces.incurs_cost import IncursCost
from sl_ai_models.resource_managers.monetary_cost_manager import MonetaryCostManager, UsagePrediction

from pydantic import BaseModel
import os
import aiohttp
import logging
logger = logging.getLogger(__name__)


class ExaSource(BaseModel, Jsonable):
    original_query: str
    auto_prompt_string: str | None
    title: str | None
    url: str | None
    text: str | None
    author: str | None
    published_date: str | None
    score: float | None
    highlights: list[str]
    highlight_scores: list[float]


class ExaSearcher(RequestLimitedModel, RetryableModel, TimeLimitedModel, IncursCost, PricedPerRequest):
    REQUESTS_PER_PERIOD_LIMIT = 25 # 25 is a guess from manual experimentation since they don't seem to publish this in an obvious place
    REQUEST_PERIOD_IN_SECONDS = 60
    TIMEOUT_TIME = 30
    PRICE_PER_REQUEST = 0.005


    async def invoke(self, search_query: str) -> list[ExaSource]:
        return await self.__retryable_timed_cost_request_limited_invoke(search_query)


    @RetryableModel._retry_according_to_model_allowed_tries
    @RequestLimitedModel._wait_till_request_capacity_available
    @IncursCost._wrap_in_cost_limiting_and_tracking
    @TimeLimitedModel._wrap_in_model_defined_timeout
    async def __retryable_timed_cost_request_limited_invoke(self, search_query: str) -> list[ExaSource]:
        response = await self._mockable_direct_call_to_model(search_query)
        return response


    async def _mockable_direct_call_to_model(self, search_query: str) -> list[ExaSource]:
        self._everything_special_to_call_before_direct_call()
        api_key = os.getenv("EXA_API_KEY")
        assert api_key is not None, "EXA_API_KEY is not set in the environment variables"
        url = "https://api.exa.ai/search"
        headers = {
            "accept": "application/json",
            "content-type": "application/json",
            "x-api-key": api_key
        }
        payload = {
            "query": search_query,
            "type": "neural",
            "useAutoprompt": True,
            "numResults": 20,
            "contents": {
            "text": {
                "includeHtmlTags": True
            },
            "highlights": {
                "numSentences": 4,
                "highlightsPerUrl": 10,
                "query": search_query
            },
            "summary": True
            }
        }

        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=payload, headers=headers) as response:
                response.raise_for_status()
                response_data: dict = await response.json()


        exa_sources: list[ExaSource] = []
        auto_prompt_string = response_data.get("autopromptString")
        for result in response_data["results"]:
            result: dict
            exa_source = ExaSource(
                original_query=search_query,
                auto_prompt_string=auto_prompt_string,
                title=result.get("title"),
                url=result.get("url"),
                text=result.get("text"),
                author=result.get("author"),
                published_date=result.get("publishedDate"),
                score=result.get("score"),
                highlights=result.get("highlights", []),
                highlight_scores=result.get("highlightScores", [])
            )
            exa_sources.append(exa_source)

        logger.debug(f"Exa API returned {len(exa_sources)} sources with urls: {[source.url for source in exa_sources]}")
        return exa_sources


    def _estimate_before_plus_after_cost_from_model_input(self, search_query: str) -> float:
        return self.PRICE_PER_REQUEST


    async def _track_cost_in_manager_using_model_response(self, response_from_direct_call: list[ExaSource], original_cost_prediction: UsagePrediction) -> None:
        assert isinstance(response_from_direct_call, list), f"response_from_direct_call is not a list, it is a {type(response_from_direct_call)}"
        cost = self.PRICE_PER_REQUEST
        MonetaryCostManager.replace_predicted_usage_with_actual_usage_in_parent_managers(original_cost_prediction, cost)


    @staticmethod
    def _get_mock_return_for_direct_call_to_model_using_cheap_input() -> list[ExaSource]:
        return [
            ExaSource(
                original_query="Moko Research Website",
                auto_prompt_string="Here is a link to the Moko Research website:",
                title="MokoResearch",
                url="https://www.mokoresearch.com",
                text="Fake text",
                author=None,
                published_date=None,
                score=0.99,
                highlights=[],
                highlight_scores=[],
            )
        ]


    @staticmethod
    def _get_cheap_input_for_invoke() -> str:
        return "MokoResearch Website"


# Example response from Exa API: (see the Exa playground for more examples)
# ... other data ... below is first source in "results" list
# {
#   "score": 0.19072401523590088,
#   "title": "Assassinated Presidents Archives - History",
#   "id": "https://www.historyonthenet.com/category/assassinated-presidents",
#   "url": "https://www.historyonthenet.com/category/assassinated-presidents",
#   "publishedDate": "2019-01-01T00:00:00.000Z",
#   "author": "None",
#   "text": "<div><div> <div> <p>Scroll down to see articles about the U.S. presidents who died in office, and the backgrounds and motivations of their assassins</p> </div> <p>Scroll down to see articles about the U.S. presidents who died in office, and the backgrounds and motivations of their assassins</p> <hr /> <article> <a href=\"https://www.historyonthenet.com/oscar-ramiro-ortega-hernandez\"> </a> <h2><a href=\"https://www.historyonthenet.com/oscar-ramiro-ortega-hernandez\">Oscar Ramiro Ortega-Hernandez</a></h2> <p>The following article on Oscar Ramiro Ortega-Hernandez is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble. Perhaps the most dangerous threat to President Obama’s life came from twenty-one-year-old Oscar Ramiro Ortega-Hernandez, who had criminal…</p> </article> <article> <a href=\"https://www.historyonthenet.com/copycat-killers\"> </a> <h2><a href=\"https://www.historyonthenet.com/copycat-killers\">Copycat Killers: Becoming Famous by Becoming Infamous</a></h2> <p>The following article on copycat killers is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble. Many assassins and would-be assassins of U.S. presidents were copycat killers obsessed with assassins from the past. Some borrowed…</p> </article> <article> <a href=\"https://www.historyonthenet.com/assassinated-presidents\"> </a> <h2><a href=\"https://www.historyonthenet.com/assassinated-presidents\">Assassinated Presidents: Profiles of Them and Their Killers</a></h2> <p>The following article on assassinated presidents is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble. The list of assassinated presidents receives a new member approximately every 20-40 years. Here are those who were killed while…</p> </article> <article> <a href=\"https://www.historyonthenet.com/isaac-aguigui\"> </a> <h2><a href=\"https://www.historyonthenet.com/isaac-aguigui\">Isaac Aguigui: Militia Leader, Wannabe Presidential Assassin</a></h2> <p>The following article on Isaac Aguigui is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble. In 2012, Barack Obama was targeted by a domestic terror group of soldiers called F.E.A.R. (“Forever Enduring Always Read”),…</p> </article> <article> <a href=\"https://www.historyonthenet.com/khalid-kelly\"> </a> <h2><a href=\"https://www.historyonthenet.com/khalid-kelly\">Khalid Kelly: Irish Would-Be Obama Assassin</a></h2> <p>The following article on Terry Kelly (\"Khalid Kelly\") is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble. In May 2011, Irish Muslim militant Terry “Khalid” Kelly was arrested for threatening to assassinate President…</p> </article> <article> <a href=\"https://www.historyonthenet.com/timothy-ryan-gutierrez-hacker-threatened-obama\"> </a> <h2><a href=\"https://www.historyonthenet.com/timothy-ryan-g",
#   "highlights": [
#     "The list of assassinated presidents receives a new member approximately every 20-40 years. Here are those who were killed while…      Isaac Aguigui: Militia Leader, Wannabe Presidential Assassin  The following article on Isaac Aguigui is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble. In 2012, Barack Obama was targeted by a domestic terror group of soldiers called F.E.A.R.",
#     "In 2012, Barack Obama was targeted by a domestic terror group of soldiers called F.E.A.R. (“Forever Enduring Always Read”),…      Khalid Kelly: Irish Would-Be Obama Assassin  The following article on Terry Kelly (\"Khalid Kelly\") is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble. In May 2011, Irish Muslim militant Terry “Khalid” Kelly was arrested for threatening to assassinate President…",
#     "Here are those who were killed while…      Isaac Aguigui: Militia Leader, Wannabe Presidential Assassin  The following article on Isaac Aguigui is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble. In 2012, Barack Obama was targeted by a domestic terror group of soldiers called F.E.A.R. (“Forever Enduring Always Read”),…      Khalid Kelly: Irish Would-Be Obama Assassin  The following article on Terry Kelly (\"Khalid Kelly\") is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama.",
#     "It is available for order now from Amazon and Barnes &amp; Noble. The list of assassinated presidents receives a new member approximately every 20-40 years. Here are those who were killed while…      Isaac Aguigui: Militia Leader, Wannabe Presidential Assassin  The following article on Isaac Aguigui is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble.",
#     "In May 2011, Irish Muslim militant Terry “Khalid” Kelly was arrested for threatening to assassinate President…",
#     "Perhaps the most dangerous threat to President Obama’s life came from twenty-one-year-old Oscar Ramiro Ortega-Hernandez, who had criminal…      Copycat Killers: Becoming Famous by Becoming Infamous  The following article on copycat killers is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble. Many assassins and would-be assassins of U.S. presidents were copycat killers obsessed with assassins from the past. Some borrowed…      Assassinated Presidents: Profiles of Them and Their Killers  The following article on assassinated presidents is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama.",
#     "It is available for order now from Amazon and Barnes &amp; Noble. In 2012, Barack Obama was targeted by a domestic terror group of soldiers called F.E.A.R. (“Forever Enduring Always Read”),…      Khalid Kelly: Irish Would-Be Obama Assassin  The following article on Terry Kelly (\"Khalid Kelly\") is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble.",
#     "Many assassins and would-be assassins of U.S. presidents were copycat killers obsessed with assassins from the past. Some borrowed…      Assassinated Presidents: Profiles of Them and Their Killers  The following article on assassinated presidents is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble. The list of assassinated presidents receives a new member approximately every 20-40 years.",
#     "It is available for order now from Amazon and Barnes &amp; Noble. Many assassins and would-be assassins of U.S. presidents were copycat killers obsessed with assassins from the past. Some borrowed…      Assassinated Presidents: Profiles of Them and Their Killers  The following article on assassinated presidents is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble.",
#     "Some borrowed…      Assassinated Presidents: Profiles of Them and Their Killers  The following article on assassinated presidents is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama. It is available for order now from Amazon and Barnes &amp; Noble. The list of assassinated presidents receives a new member approximately every 20-40 years. Here are those who were killed while…      Isaac Aguigui: Militia Leader, Wannabe Presidential Assassin  The following article on Isaac Aguigui is an excerpt from Mel Ayton's Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama."
#   ],
#   "highlightScores": [
#     0.10017715394496918,
#     0.0916084423661232,
#     0.08568621426820755,
#     0.08466880768537521,
#     0.08453669399023056,
#     0.08243578672409058,
#     0.08049978315830231,
#     0.08013768494129181,
#     0.0784364566206932,
#     0.07647785544395447
#   ],
#   "summary": "This webpage is a collection of articles about U.S. presidents who died in office, and the backgrounds and motivations of their assassins. It includes excerpts from the book \"Hunting the President: Threats, Plots, and Assassination Attempts—From FDR to Obama\" by Mel Ayton. The articles cover a variety of topics, including the assassination attempt on President Obama by Oscar Ramiro Ortega-Hernandez, the motivations of copycat killers, and the various individuals who have attempted to assassinate U.S. presidents throughout history. \n"
# }


from sl_ai_models.basic_model_interfaces.ai_model import AiModel
import asyncio
import pytest
from tests.utilities_for_tests import coroutine_testing
from tests.models_to_test import ModelsToTest
from tests.ai_mock_manager import AiModelMockManager
from unittest.mock import Mock
import os
import logging
logger = logging.getLogger(__name__)
from sl_ai_models.perplexity import Perplexity
from sl_ai_models.exa_searcher import ExaSearcher
from sl_ai_models.deepgram_nova2 import DeepgramNova2

@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.BASIC_MODEL_LIST
)
def test_response_from_a_direct_call_is_same_ask_mock_value(subclass: type[AiModel]) -> None:
    if subclass in [Perplexity, ExaSearcher, DeepgramNova2]:
        pytest.skip(f"Skipping {subclass.__name__} return value is inconsistent")

    model = subclass()
    model_input = model._get_cheap_input_for_invoke()
    response = asyncio.run(model._mockable_direct_call_to_model(model_input))
    assert response is not None, "Response is None"

    mock_value = model._get_mock_return_for_direct_call_to_model_using_cheap_input()
    logger.info(f"Response: {response}, Mock Value: {mock_value}")

    assert response == mock_value, f"Response is not the same as the mock value. Response: {response}, Mock Value: {mock_value}"


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.BASIC_MODEL_LIST
)
def test_ai_model_returns_response_with_invoke(subclass: type[AiModel]) -> None:
    model = subclass()
    model_input = model._get_cheap_input_for_invoke()
    response = asyncio.run(model.invoke(model_input))
    assert response is not None, "Response is None"



@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.BASIC_MODEL_LIST
)
def test_ai_model_async_is_not_blocking(subclass: type[AiModel]) -> None:
    # NOTE: Don't mock this unless costs get bad since blocking could be caused at any code level below the mock
    number_of_coroutines_to_run = 5
    list_should_run_under_x_times_first_coroutine = 3

    model_input = subclass._get_cheap_input_for_invoke()
    first_coroutine = subclass().invoke(model_input)
    list_of_coroutines = [subclass().invoke(model_input) for _ in range(number_of_coroutines_to_run)]
    coroutine_testing.assert_coroutines_run_under_x_times_duration_of_benchmark([first_coroutine], list_of_coroutines, x=list_should_run_under_x_times_first_coroutine, allowed_errors_or_timeouts=5)


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.BASIC_MODEL_LIST
)
def test_call_limit_error_not_raise_if_not_in_test_env(mocker: Mock, subclass: type[AiModel]) -> None:
    pytest_current_test = os.environ.get('PYTEST_CURRENT_TEST')
    if pytest_current_test is None:
        raise RuntimeError("PYTEST_CURRENT_TEST is not set")
    try:
        del os.environ['PYTEST_CURRENT_TEST']
        model = subclass()
        max_calls = 1

        model._increment_calls_then_error_if_testing_call_limit_reached(max_calls)
        model._increment_calls_then_error_if_testing_call_limit_reached(max_calls)
        model._increment_calls_then_error_if_testing_call_limit_reached(max_calls)
        os.environ['PYTEST_CURRENT_TEST'] = pytest_current_test
    except Exception as e:
        os.environ['PYTEST_CURRENT_TEST'] = pytest_current_test
        assert False, f"Error: {e}"


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.BASIC_MODEL_LIST
)
def test_call_limit_error_raised_so_tests_cant_accidentally_create_insane_costs(subclass: type[AiModel]) -> None:
    model = subclass()
    calls_before = model._num_calls_to_dependent_model.get()
    try:
        model._num_calls_to_dependent_model.set(0)
        max_calls = 1
        model._increment_calls_then_error_if_testing_call_limit_reached(max_calls)

        with pytest.raises(RuntimeError):
            model._increment_calls_then_error_if_testing_call_limit_reached(max_calls)
    finally:
        model._num_calls_to_dependent_model.set(calls_before)

@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.BASIC_MODEL_LIST
)
def test_langchain_traces_deactivated_if_in_testing_mode(mocker: Mock, subclass: type[AiModel]) -> None:
    model = subclass()
    model._deactivate_langchain_traces_if_in_testing_mode()
    langchain_api_key = os.environ.get('LANGCHAIN_API_KEY')
    tracing_v2 = os.environ.get('LANGCHAIN_TRACING_V2')
    assert langchain_api_key is None, "Langchain traces were not deactivated"
    assert tracing_v2 is None, "Langchain traces were not deactivated"


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.BASIC_MODEL_LIST
)
def test_special_functions_called_with_direct_call(mocker: Mock, subclass: type[AiModel]):
    model = subclass()
    mock_limiting_function = AiModelMockManager.mock_function_that_throws_error_if_test_limit_reached(mocker)
    mock_langchain_deactivating_function = AiModelMockManager.mock_function_that_deactivates_langchain_tracing(mocker)
    asyncio.run(model._mockable_direct_call_to_model(model._get_cheap_input_for_invoke()))
    limit_function_num_calls = mock_limiting_function.call_count
    langchain_function_num_calls = mock_langchain_deactivating_function.call_count
    assert limit_function_num_calls == 1, f"Model function was not called n=1 times. It was called {limit_function_num_calls} times"
    assert langchain_function_num_calls == 1, f"Langchain function was not called n=1 times. It was called {langchain_function_num_calls} times"


from typing import Final
from sl_ai_models.model_archetypes.openai_vision_model import OpenAiVisionToTextModel
from sl_ai_models.utils.openai_utils import VisionMessageData
from sl_ai_models.utils.ai_misc import clean_indents  # Keep this import here for easier imports into other files so prompts can keep proper indentation levels in code # NOSONAR
from typing import TypeVar
T = TypeVar("T")

class Gpt4VisionInput(VisionMessageData):
    # This class is just to allow additional fields later, and easier imports so you can import all you need from one file
    pass


class Gpt4oVision(OpenAiVisionToTextModel):
    MODEL_NAME: Final[str] = "gpt-4o"
    REQUESTS_PER_PERIOD_LIMIT: Final[int] = 4000 # Errors said the limit is 4k, but it says 100k online See OpenAI Limit on the account dashboard for most up-to-date limit
    REQUEST_PERIOD_IN_SECONDS: Final[int] = 60
    TIMEOUT_TIME: Final[int] = 40
    TOKENS_PER_PERIOD_LIMIT: Final[int] = 35000 # Actual rate is 40k, but lowered for wiggle room. See OpenAI Limit on the account dashboard for most up-to-date limit
    TOKEN_PERIOD_IN_SECONDS: Final[int] = 60





from sl_ai_models.gpt4ovision import Gpt4oVision
from sl_ai_models.gpt4o import Gpt4o
from sl_ai_models.perplexity import Perplexity
from sl_ai_models.claude35sonnet import Claude35Sonnet
from sl_ai_models.exa_searcher import ExaSearcher
from sl_ai_models.gpt4omini import Gpt4oMini
from sl_ai_models.deepgram_nova2 import DeepgramNova2
from sl_ai_models.claude3opus import Claude3Opus

from sl_ai_models.resource_managers.monetary_cost_manager import MonetaryCostManager

from sl_ai_models.utils.async_batching import run_coroutines, run_coroutines_while_removing_and_logging_exceptions, wrap_coroutines_to_return_not_raise_exceptions
from sl_ai_models.utils.ai_misc import clean_indents


from typing import Final
from sl_ai_models.model_archetypes.anthropic_text_model import AnthropicTextToTextModel
from sl_ai_models.utils.ai_misc import clean_indents  # Keep this import here for easier imports into other files so prompts can keep proper indentation levels in code # NOSONAR
from typing import TypeVar
T = TypeVar("T")

class Claude35Sonnet(AnthropicTextToTextModel):
    # See Anthropic Limit on the account dashboard for most up-to-date limit
    MODEL_NAME: Final[str] = "claude-3-5-sonnet-20240620"
    REQUESTS_PER_PERIOD_LIMIT: Final[int] = 50
    REQUEST_PERIOD_IN_SECONDS: Final[int] = 60
    TIMEOUT_TIME: Final[int] = 40
    TOKENS_PER_PERIOD_LIMIT: Final[int] = 40000
    TOKEN_PERIOD_IN_SECONDS: Final[int] = 60




import pytest
from tests.models_to_test import ModelsToTest
from tests.ai_mock_manager import AiModelMockManager
from sl_ai_models.basic_model_interfaces.token_limited_model import TokenLimitedModel
from unittest.mock import Mock
from sl_ai_models.basic_model_interfaces.ai_model import AiModel
import logging
logger = logging.getLogger(__name__)
from sl_ai_models.utils import async_batching
import asyncio

TOKEN_LIMITED_ERROR_MESSAGE = "Model must be TokenLimited"
MOCK_INPUT_TO_TOKEN_RETURN_VALUE = 5000

# TODO: Also check that the burst works by itself

@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.TOKEN_LIMITED_LIST
)
def test_token_amount_past_burst_doesnt_happen_instantly(mocker: Mock, subclass: type[AiModel]) -> None:
    if not issubclass(subclass, TokenLimitedModel):
        raise ValueError(TOKEN_LIMITED_ERROR_MESSAGE)


    AiModelMockManager.mock_ai_model_direct_call_with_predefined_mock_value(mocker, subclass)
    AiModelMockManager.reinitialize_limiters(subclass)

    model = subclass()
    invoke_input = subclass._get_cheap_input_for_invoke()
    tokens_to_reach_burst = get_number_of_tokens_to_deplete_burst(subclass)
    tokens_after_burst = get_number_of_tokens_that_takes_10s_to_run_given_model_rate(subclass)
    total_tokens = tokens_to_reach_burst + tokens_after_burst
    number_of_calls_to_make = mock_input_to_tokens_and_get_call_count(mocker, subclass, total_tokens)
    coroutines_to_test = [model.invoke(invoke_input) for _ in range(number_of_calls_to_make)]

    with pytest.raises(asyncio.TimeoutError):
        seconds_coroutines_should_take_longer_than = 5
        timed_coroutines = async_batching.wrap_coroutines_with_timeout(coroutines_to_test, seconds_coroutines_should_take_longer_than)
        async_batching.run_coroutines(timed_coroutines)



def get_number_of_tokens_to_deplete_burst(subclass: type[TokenLimitedModel]) -> int:
    tokens_to_reach_burst = subclass.TOKENS_PER_PERIOD_LIMIT
    return tokens_to_reach_burst


def get_number_of_tokens_that_takes_10s_to_run_given_model_rate(subclass: type[TokenLimitedModel]) -> int:
    time_tests_should_take_in_seconds = 10
    allowed_tokens_per_second = subclass.TOKENS_PER_PERIOD_LIMIT / subclass.TOKEN_PERIOD_IN_SECONDS
    number_of_tokens_to_ask_for = int(time_tests_should_take_in_seconds * allowed_tokens_per_second)
    return number_of_tokens_to_ask_for


def mock_input_to_tokens_and_get_call_count(mocker:Mock, subclass: type[TokenLimitedModel], tokens: int) -> int:
    mock_input_to_token_return_value = int(subclass.TOKENS_PER_PERIOD_LIMIT / 1000)

    AiModelMockManager.mock_input_to_tokens_with_value(mocker, subclass, mock_input_to_token_return_value)

    number_of_calls_to_make = tokens // mock_input_to_token_return_value
    return number_of_calls_to_make


from typing import Any
from abc import ABC
from deepgram import DeepgramClient, PrerecordedOptions, PrerecordedResponse
import os
from pydantic import BaseModel
from sl_ai_models.basic_model_interfaces.named_model import NamedModel
from sl_ai_models.basic_model_interfaces.time_limited_model import TimeLimitedModel
from sl_ai_models.basic_model_interfaces.request_limited_model import RequestLimitedModel
from sl_ai_models.basic_model_interfaces.retryable_model import RetryableModel
import logging
import aiofiles
logger = logging.getLogger(__name__)


class DeepgramOutput(BaseModel):
    transcript: str
    audio_duration_in_seconds: float
    other_data: dict | None = None


class DeepgramAudioToTextModel(
    NamedModel,
    TimeLimitedModel,
    RequestLimitedModel,
    RetryableModel,
    ABC
    ):

    async def invoke_and_return_just_transcript(self, url: str) -> str:
        output = await self.invoke(url)
        return output.transcript


    async def invoke(self, url_or_absolute_file_path: str) -> DeepgramOutput:
        logger.debug(f"Invoking Deepgram model with url: {url_or_absolute_file_path}")
        result = await self.__invoke_with_retry_request_and_time_limits(url_or_absolute_file_path)
        logger.debug(f"Successfully invoked Deepgram model with url: {url_or_absolute_file_path}")
        return result


    @RetryableModel._retry_according_to_model_allowed_tries
    @RequestLimitedModel._wait_till_request_capacity_available
    @TimeLimitedModel._wrap_in_model_defined_timeout
    async def __invoke_with_retry_request_and_time_limits(self, url_or_absolute_file_path: str) -> DeepgramOutput:
        return await self._mockable_direct_call_to_model(url_or_absolute_file_path)


    async def _mockable_direct_call_to_model(self, url_or_absolute_file_path: str) -> DeepgramOutput:
        self._everything_special_to_call_before_direct_call()
        response: PrerecordedResponse = await self.__call_deepgram_online_model(url_or_absolute_file_path)
        output = self.__turn_deepgram_prerecorded_response_into_output(response)
        return output


    async def __call_deepgram_online_model(self, url_or_absolute_file_path: str) -> PrerecordedResponse:
        deepgram_api_key = os.environ.get("DEEPGRAM_API_KEY")
        assert deepgram_api_key is not None
        deepgram = DeepgramClient(deepgram_api_key)
        options = PrerecordedOptions(
            smart_format=True,
            model=self.MODEL_NAME,
            language="en-US",
            diarize=True,
            filler_words=False,
        )

        listener = deepgram.listen.asyncrest.v("1")
        time_out_time = 300
        if url_or_absolute_file_path.startswith("http"):
            payload = {
                'url': url_or_absolute_file_path
            }
            response = await listener.transcribe_url(payload, options, timeout=time_out_time) # type: ignore
        else:
            async with aiofiles.open(url_or_absolute_file_path, "rb") as file:
                buffer_data = await file.read()
            payload = { 'buffer': buffer_data }
            response = await listener.transcribe_file(payload, options, timeout=time_out_time) # type: ignore

        assert isinstance(response, PrerecordedResponse)
        return response


    def __turn_deepgram_prerecorded_response_into_output(self, response: PrerecordedResponse) -> DeepgramOutput:
        response_json: dict = response.to_dict()
        try:
            transcript = response_json["results"]["channels"][0]["alternatives"][0]["paragraphs"]["transcript"] # type: ignore # Any of these things can be None, but its too messy to assert on all of them
            audio_duration = float(response_json["metadata"]["duration"]) # type: ignore
        except Exception as e:
            raise ValueError(f"Error getting transcript from Deepgram response. Error: {e}. Response: {response_json}")
        output = DeepgramOutput(
            transcript=transcript,
            audio_duration_in_seconds=audio_duration,
            other_data=response_json,
            )
        return output


    @staticmethod
    def _get_mock_return_for_direct_call_to_model_using_cheap_input() -> Any:
        return MOCK_RETURN_FOR_DIRECT_CALL_TO_MODEL


    @staticmethod
    def _get_cheap_input_for_invoke() -> Any:
        return "https://static.deepgram.com/examples/Bueller-Life-moves-pretty-fast.wav"



MOCK_RETURN_FOR_DIRECT_CALL_TO_MODEL = DeepgramOutput(
    transcript="\nSpeaker 0: Yep. I said it before, and I'll say it again. Life moves pretty fast. You don't stop and look around once in a while, you could miss it.",
    audio_duration_in_seconds=17.566313,
    other_data={
        "metadata": {
            "transaction_key": "deprecated",
            "request_id": "33080039-be8e-45b7-a5e0-5fd91cff5cf9",
            "sha256": "5324da68ede209a16ac69a38e8cd29cee4d754434a041166cda3a1f5e0b24566",
            "created": "2024-08-16T22:19:30.446Z",
            "duration": 17.566313,
            "channels": 1,
            "models": [
                "30089e05-99d1-4376-b32e-c263170674af"
            ],
            "model_info": {
                "30089e05-99d1-4376-b32e-c263170674af": {
                    "name": "2-general-nova",
                    "version": "2024-01-09.29447",
                    "arch": "nova-2"
                }
            }
        },
        "results": {
            "channels": [
                {
                    "alternatives": [
                        {
                            "transcript": "Yep. I said it before, and I'll say it again. Life moves pretty fast. You don't stop and look around once in a while, you could miss it.",
                            "confidence": 0.99872786,
                            "words": [
                                {
                                    "word": "yep",
                                    "start": 5.52,
                                    "end": 6.02,
                                    "confidence": 0.9958389,
                                    "punctuated_word": "Yep.",
                                    "speaker": 0,
                                    "speaker_confidence": 0.9121094
                                },
                                {
                                    "word": "i",
                                    "start": 7.04,
                                    "end": 7.2799997,
                                    "confidence": 0.512589,
                                    "punctuated_word": "I",
                                    "speaker": 0,
                                    "speaker_confidence": 0.9121094
                                },
                                {
                                    "word": "said",
                                    "start": 7.2799997,
                                    "end": 7.44,
                                    "confidence": 0.9672884,
                                    "punctuated_word": "said",
                                    "speaker": 0,
                                    "speaker_confidence": 0.9121094
                                },
                                {
                                    "word": "it",
                                    "start": 7.44,
                                    "end": 7.6,
                                    "confidence": 0.9997284,
                                    "punctuated_word": "it",
                                    "speaker": 0,
                                    "speaker_confidence": 0.9121094
                                },
                                {
                                    "word": "before",
                                    "start": 7.6,
                                    "end": 7.9199996,
                                    "confidence": 0.7845876,
                                    "punctuated_word": "before,",
                                    "speaker": 0,
                                    "speaker_confidence": 0.9121094
                                },
                                {
                                    "word": "and",
                                    "start": 7.9199996,
                                    "end": 8.08,
                                    "confidence": 0.9998627,
                                    "punctuated_word": "and",
                                    "speaker": 0,
                                    "speaker_confidence": 0.9121094
                                },
                                {
                                    "word": "I'll", "start": 8.08, "end": 8.32, "confidence": 0.99989426, "punctuated_word": "I'll",
                                    "speaker": 0,
                                    "speaker_confidence": 0.9121094
                                },
                                {
                                    "word": "say",
                                    "start": 8.32,
                                    "end": 8.48,
                                    "confidence": 0.9996518,
                                    "punctuated_word": "say",
                                    "speaker": 0,
                                    "speaker_confidence": 0.9121094
                                },
                                {
                                    "word": "it",
                                    "start": 8.48,
                                    "end": 8.639999,
                                    "confidence": 0.9998282,
                                    "punctuated_word": "it",
                                    "speaker": 0,
                                    "speaker_confidence": 0.92822266
                                },
                                {
                                    "word": "again",
                                    "start": 8.639999,
                                    "end": 9.139999,
                                    "confidence": 0.9737228,
                                    "punctuated_word": "again.",
                                    "speaker": 0,
                                    "speaker_confidence": 0.92822266
                                },
                                {
                                    "word": "life",
                                    "start": 9.991312,
                                    "end": 10.391313,
                                    "confidence": 0.99582124,
                                    "punctuated_word": "Life",
                                    "speaker": 0,
                                    "speaker_confidence": 0.92822266
                                },
                                {
                                    "word": "moves",
                                    "start": 10.391313,
                                    "end": 10.711312,
                                    "confidence": 0.9989655,
                                    "punctuated_word": "moves",
                                    "speaker": 0,
                                    "speaker_confidence": 0.92822266
                                },
                                {
                                    "word": "pretty",
                                    "start": 10.711312,
                                    "end": 11.031313,
                                    "confidence": 0.9996259,
                                    "punctuated_word": "pretty",
                                    "speaker": 0,
                                    "speaker_confidence": 0.92822266
                                },
                                {
                                    "word": "fast",
                                    "start": 11.031313,
                                    "end": 11.531313,
                                    "confidence": 0.99960166,
                                    "punctuated_word": "fast.",
                                    "speaker": 0,
                                    "speaker_confidence": 0.92822266
                                },
                                {
                                    "word": "you",
                                    "start": 12.071312,
                                    "end": 12.231313,
                                    "confidence": 0.950412,
                                    "punctuated_word": "You",
                                    "speaker": 0,
                                    "speaker_confidence": 0.8408203
                                },
                                {
                                    "word": "don't", "start": 12.231313, "end": 12.4713125, "confidence": 0.9998907, "punctuated_word": "don't",
                                    "speaker": 0,
                                    "speaker_confidence": 0.8408203
                                },
                                {
                                    "word": "stop",
                                    "start": 12.4713125,
                                    "end": 12.711312,
                                    "confidence": 0.99980956,
                                    "punctuated_word": "stop",
                                    "speaker": 0,
                                    "speaker_confidence": 0.8408203
                                },
                                {
                                    "word": "and",
                                    "start": 12.711312,
                                    "end": 12.871312,
                                    "confidence": 0.99874026,
                                    "punctuated_word": "and",
                                    "speaker": 0,
                                    "speaker_confidence": 0.8408203
                                },
                                {
                                    "word": "look",
                                    "start": 12.871312,
                                    "end": 13.031313,
                                    "confidence": 0.99971575,
                                    "punctuated_word": "look",
                                    "speaker": 0,
                                    "speaker_confidence": 0.8408203
                                },
                                {
                                    "word": "around",
                                    "start": 13.031313,
                                    "end": 13.351313,
                                    "confidence": 0.99964535,
                                    "punctuated_word": "around",
                                    "speaker": 0,
                                    "speaker_confidence": 0.8408203
                                },
                                {
                                    "word": "once",
                                    "start": 13.351313,
                                    "end": 13.591312,
                                    "confidence": 0.99837434,
                                    "punctuated_word": "once",
                                    "speaker": 0,
                                    "speaker_confidence": 0.8408203
                                },
                                {
                                    "word": "in",
                                    "start": 13.591312,
                                    "end": 13.671312,
                                    "confidence": 0.9986059,
                                    "punctuated_word": "in",
                                    "speaker": 0,
                                    "speaker_confidence": 0.8408203
                                },
                                {
                                    "word": "a",
                                    "start": 13.671312,
                                    "end": 13.831312,
                                    "confidence": 0.9862984,
                                    "punctuated_word": "a",
                                    "speaker": 0,
                                    "speaker_confidence": 0.94873047
                                },
                                {
                                    "word": "while",
                                    "start": 13.831312,
                                    "end": 14.331312,
                                    "confidence": 0.93138945,
                                    "punctuated_word": "while,",
                                    "speaker": 0,
                                    "speaker_confidence": 0.94873047
                                },
                                {
                                    "word": "you",
                                    "start": 14.631312,
                                    "end": 14.791312,
                                    "confidence": 0.9974233,
                                    "punctuated_word": "you",
                                    "speaker": 0,
                                    "speaker_confidence": 0.94873047
                                },
                                {
                                    "word": "could",
                                    "start": 14.791312,
                                    "end": 14.951313,
                                    "confidence": 0.9986802,
                                    "punctuated_word": "could",
                                    "speaker": 0,
                                    "speaker_confidence": 0.94873047
                                },
                                {
                                    "word": "miss",
                                    "start": 14.951313,
                                    "end": 15.191313,
                                    "confidence": 0.99872786,
                                    "punctuated_word": "miss",
                                    "speaker": 0,
                                    "speaker_confidence": 0.94873047
                                },
                                {
                                    "word": "it",
                                    "start": 15.191313,
                                    "end": 15.691313,
                                    "confidence": 0.99549305,
                                    "punctuated_word": "it.",
                                    "speaker": 0,
                                    "speaker_confidence": 0.94873047
                                }
                            ],
                            "paragraphs": {
                                "transcript": "\nSpeaker 0: Yep. I said it before, and I'll say it again. Life moves pretty fast. You don't stop and look around once in a while, you could miss it.",
                                "paragraphs": [
                                    {
                                        "sentences": [
                                            {
                                                "text": "Yep.",
                                                "start": 5.52,
                                                "end": 6.02
                                            },
                                            {
                                                "text": "I said it before, and I'll say it again.", "start": 7.04, "end": 9.139999}, {"text": "Life moves pretty fast.", "start": 9.991312, "end": 11.531313}, {"text": "You don't stop and look around once in a while, you could miss it.",
                                                "start": 12.071312,
                                                "end": 15.691313
                                            }
                                        ],
                                        "start": 5.52,
                                        "end": 15.691313,
                                        "num_words": 28,
                                        "speaker": 0
                                    }
                                ]
                            }
                        }
                    ]
                }
            ]
        }
    }
)






from sl_ai_models.model_archetypes.deepgram_model import DeepgramAudioToTextModel, DeepgramOutput #NOSONAR # Keep DeepgramOutput here for easier imports
from typing import Final

class DeepgramNova2(DeepgramAudioToTextModel):
    MODEL_NAME: Final[str] = "nova-2"
    TIMEOUT_TIME: Final[int] = 85
    REQUESTS_PER_PERIOD_LIMIT: Final[int] = 100 # Technically its that you can only have 100 concurrent requests, but its easier to reuse rate limiting https://developers.deepgram.com/reference/api-rate-limits
    REQUEST_PERIOD_IN_SECONDS: Final[int] = 90



# This file is run before any tests are run in order to configure tests

import pytest
from sl_ai_models.utils.custom_logger import CustomLogger

@pytest.fixture(scope="session", autouse=True)
def setup_logging() -> None:
    CustomLogger.set_up_logging()


from __future__ import annotations
from typing import Final
import logging
logger = logging.getLogger(__name__)
from contextvars import ContextVar


class HardLimitExceededError(Exception):
    """Raised when the hardlimit is exceeded"""


class HardLimitManager:
    _active_limit_managers: ContextVar[list[HardLimitManager]] = ContextVar('_active_limit_managers', default=[])
    _id_counter: int = 0

    def __init__(self, hard_limit: float, log_usage_when_called: bool = False) -> None:
        self.hard_limit: Final[float] = hard_limit
        self._current_usage: float = 0
        self.__log_usage_when_called: bool = log_usage_when_called
        HardLimitManager._id_counter += 1
        self.id = HardLimitManager._id_counter


    @property
    def current_usage(self) -> float:
        return self._current_usage


    @property
    def amount_left(self) -> float:
        return self.hard_limit - self._current_usage


    @classmethod
    def get_active_cost_managers(cls) -> list[HardLimitManager]:
        return cls._active_limit_managers.get()


    def __enter__(self) -> HardLimitManager:
        cost_managers = self._active_limit_managers.get().copy()
        cost_managers.append(self)
        self._active_limit_managers.set(cost_managers)
        return self


    def __exit__(self, exc_type, exc_value, traceback) -> None: #NOSONAR
        # This stuff might be redundant
        cost_managers = self._active_limit_managers.get()
        cost_managers.remove(self)
        self._active_limit_managers.set(cost_managers)


    @classmethod
    def raise_error_if_limit_would_be_reached(cls, amount_to_check_room_for: float = 0) -> None:
        """
        You can put in 0 for the amount to check if the current usage is at the limit
        """
        if amount_to_check_room_for < 0:
            raise ValueError("Amount should be a positive number or zero")
        for cost_manager in cls._active_limit_managers.get():
            if cost_manager.amount_left < amount_to_check_room_for:
                raise HardLimitExceededError(f"Usage amount {amount_to_check_room_for} would push current usage to {cost_manager.current_usage + amount_to_check_room_for} exceeding the hard limit of {cost_manager.hard_limit}")


    @classmethod
    def increase_current_usage_in_parent_managers(cls, amount: float) -> None:
        if amount < 0:
            raise ValueError("Cost should be a positive number or zero")
        if amount == 0:
            logger.info("The cost inputted is zero which may or may not be a problem")
        for cost_manager in cls._active_limit_managers.get():
            cost_manager._current_usage += amount
            if cost_manager._current_usage > cost_manager.hard_limit:
                logger.warn(f"Usage increase exceeded the hard limit of {cost_manager.hard_limit}")
            if cost_manager.__log_usage_when_called:
                logger.info(f"{cost_manager.__class__}.ID{cost_manager.id}. Current usage now {cost_manager._current_usage}. Cost of {amount} added")







from __future__ import annotations
from abc import ABC
from sl_ai_models.utils import file_manipulation
import logging
from pydantic import BaseModel
import json
from typing import Any, TypeVar
logger = logging.getLogger(__name__)

T = TypeVar('T', bound='Jsonable')

class Jsonable(ABC):
    """
    An interface that allows a class to be converted to and from json
    """

    def to_json(self) -> dict:
        if isinstance(self, BaseModel):
            return self._pydantic_model_to_dict(self)
        else:
            raise NotImplementedError(f"Class {self.__class__.__name__} does not have a to_json method.")


    @classmethod
    def from_json(cls: type[T], json: dict) -> T:
        if issubclass(cls, BaseModel):
            return cls._pydantic_model_from_dict(cls, json)
        else:
            raise NotImplementedError(f"Class {cls.__name__} does not have a from_json method. This should be implemented in the subclass.")


    @classmethod
    def convert_project_file_path_to_object_list(cls: type[T], project_file_path: str) -> list[T]:
        return cls._use__from_json__to_convert_project_file_path_to_object_list(project_file_path)


    @classmethod
    def _use__from_json__to_convert_project_file_path_to_object_list(cls: type[T], project_file_path: str) -> list[T]:
        jsons = file_manipulation.load_json_file(project_file_path)
        assert isinstance(jsons, list), f"The json file at {project_file_path} did not contain a list."
        objects = [cls.from_json(json) for json in jsons]
        return objects


    @staticmethod
    def save_object_list_to_file_path(objects: list[Jsonable], file_path_from_top_of_project: str) -> None:
        file_manipulation.write_json_file(file_path_from_top_of_project, [object.to_json() for object in objects])


    @staticmethod
    def _pydantic_model_to_dict(pydantic_model: BaseModel) -> dict:
        json_string: str = pydantic_model.model_dump_json()
        json_dict: dict = json.loads(json_string)
        return json_dict


    @staticmethod
    def _pydantic_model_from_dict(cls_type: type[BaseModel], json_dict: dict) -> Any:
        json_string: str = json.dumps(json_dict)
        return cls_type.model_validate_json(json_string)





from tests.models_to_test import ModelsToTest
from tests.ai_mock_manager import AiModelMockManager
from sl_ai_models.resource_managers.monetary_cost_manager import MonetaryCostManager, HardLimitExceededError
import pytest
from unittest.mock import Mock
from sl_ai_models.basic_model_interfaces.ai_model import AiModel
from sl_ai_models.basic_model_interfaces.incurs_cost import IncursCost
import asyncio
from sl_ai_models.utils import async_batching
import logging
logger = logging.getLogger(__name__)
import random
from sl_ai_models.utils.response_types import TextTokenCostResponse
from sl_ai_models.basic_model_interfaces.request_limited_model import RequestLimitedModel


############################### HELPERS ########################################
NOT_INCURS_COST_ERROR_MESSAGE = "Model must have interface IncursCost"

async def async_run_cheap_invoke(subclass: type[AiModel]) -> None:
    model = subclass()
    model_input = model._get_cheap_input_for_invoke()
    random_time_to_wait = random.uniform(0.05, 0.1)
    await asyncio.sleep(random_time_to_wait)
    await model.invoke(model_input)


def run_cheap_invoke(subclass: type[AiModel]) -> None:
    asyncio.run(async_run_cheap_invoke(subclass))


def run_cheap_invoke_and_track_cost(subclass: type[AiModel], max_cost: float) -> float:
    running_cost: float = 0
    with MonetaryCostManager(max_cost) as cost_manager:
        run_cheap_invoke(subclass)
        running_cost = cost_manager.current_usage

    return running_cost


def find_number_of_calls_to_make(subclass: type[AiModel]):
    if issubclass(subclass, RequestLimitedModel) and subclass.REQUESTS_PER_PERIOD_LIMIT < 100:
        return 10
    else:
        return 100


def find_number_of_hard_limit_exceptions_in_run(mocker: Mock, subclass: type[AiModel], max_cost: float, number_of_calls_to_make: int, number_of_expected_exceptions: int) -> int:
    AiModelMockManager.mock_ai_model_direct_call_with_predefined_mock_value(mocker, subclass)

    coroutines = [async_run_cheap_invoke(subclass) for _ in range(number_of_calls_to_make)]
    exception_handled_coroutines = async_batching.wrap_coroutines_to_return_not_raise_exceptions(coroutines)

    with MonetaryCostManager(max_cost):
        results = async_batching.run_coroutines(exception_handled_coroutines)
        logger.info(results)

    number_of_exceptions = len([result for result in results if isinstance(result, Exception)])
    number_of_hard_limit_exceptions = len([result for result in results if isinstance(result, HardLimitExceededError)])
    assert number_of_hard_limit_exceptions == number_of_exceptions, "Some exceptions were not HardLimitExceededError"

    return number_of_hard_limit_exceptions


############################### TESTS ########################################

@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.INCURS_COST_LIST
)
def test_cost_manager_notices_cost_without_mocks(subclass: type[AiModel]) -> None:
    if not issubclass(subclass, IncursCost):
        raise ValueError(NOT_INCURS_COST_ERROR_MESSAGE)

    max_cost = 100
    cost = run_cheap_invoke_and_track_cost(subclass, max_cost)
    assert cost > 0, "No cost was incurred"


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.INCURS_COST_LIST
)
def test_cost_manager_notices_cost_with_mocks(mocker: Mock, subclass: type[AiModel]) -> None:
    if not issubclass(subclass, IncursCost):
        raise ValueError(NOT_INCURS_COST_ERROR_MESSAGE)

    AiModelMockManager.mock_ai_model_direct_call_with_predefined_mock_value(mocker, subclass)
    max_cost = 100
    cost = run_cheap_invoke_and_track_cost(subclass, max_cost)
    assert cost > 0, "No cost was incurred"


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.INCURS_COST_LIST
)
def test_error_thrown_when_limit_reached(mocker: Mock, subclass: type[AiModel]) -> None:
    if not issubclass(subclass, IncursCost):
        raise ValueError(NOT_INCURS_COST_ERROR_MESSAGE)

    AiModelMockManager.mock_ai_model_direct_call_with_predefined_mock_value(mocker, subclass)
    max_cost = 0.0000000000001
    with MonetaryCostManager(max_cost):
        with pytest.raises(HardLimitExceededError):
            run_cheap_invoke(subclass) # This first line might not need to raise an error, since the cost is not exceeded before the calls is made. Depends on current cost implementation
            run_cheap_invoke(subclass)


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.INCURS_COST_LIST
)
def test_error_thrown_with_many_calls_and_limit_of_0(mocker: Mock, subclass: type[AiModel]) -> None:
    if not issubclass(subclass, IncursCost):
        raise ValueError(NOT_INCURS_COST_ERROR_MESSAGE)

    max_cost = 0.0000000000001
    number_of_calls_to_make = find_number_of_calls_to_make(subclass)
    number_of_expected_exceptions = number_of_calls_to_make - 1
    num_hard_limit_errors = find_number_of_hard_limit_exceptions_in_run(mocker, subclass, max_cost, number_of_calls_to_make, number_of_expected_exceptions)
    assert num_hard_limit_errors > number_of_expected_exceptions, "Improper number of exceptions thrown. Only first call should be allowed to possibly exceed since cost would be exceeded after this"


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.INCURS_COST_LIST
)
def test_error_thrown_when_many_calls_with_a_larger_than_0_cost_limit(mocker: Mock, subclass: type[AiModel]) -> None:
    if not issubclass(subclass, IncursCost):
        raise ValueError(NOT_INCURS_COST_ERROR_MESSAGE)

    number_of_calls_to_make = find_number_of_calls_to_make(subclass)
    mock_return_value = subclass._get_mock_return_for_direct_call_to_model_using_cheap_input()
    if isinstance(mock_return_value, TextTokenCostResponse):
        cost_of_mock_call = mock_return_value.cost
    else:
        cost_of_mock_call = 0.005 # Estimate

    percentage_we_want_to_error = 0.5
    max_cost = cost_of_mock_call * number_of_calls_to_make * percentage_we_want_to_error
    number_of_expected_exceptions_given_perfect_cost_prediction = int(number_of_calls_to_make * percentage_we_want_to_error) -1

    # Especially with the low token amount of the cheap input, its likely that the completion prediction will underestimate the true completion token amount, and thus not error enough.
    allowed_percentage_to_be_off_by = 0.5
    number_of_expected_exceptions = number_of_expected_exceptions_given_perfect_cost_prediction * allowed_percentage_to_be_off_by

    num_hard_limit_errors = find_number_of_hard_limit_exceptions_in_run(mocker, subclass, max_cost, number_of_calls_to_make, number_of_expected_exceptions_given_perfect_cost_prediction)
    assert num_hard_limit_errors > number_of_expected_exceptions, "The cost manager let too many calls through"

    number_of_calls_that_did_not_error = number_of_calls_to_make - num_hard_limit_errors
    assert number_of_calls_that_did_not_error > 1, "Only one call was allowed through"


import asyncio
import pytest
from unittest.mock import Mock
from typing import Coroutine, Callable
import time
import logging
logger = logging.getLogger(__name__)
from enum import Enum


from tests.ai_mock_manager import AiModelMockManager
from tests.models_to_test import ModelsToTest
from tests.utilities_for_tests import coroutine_testing
from sl_ai_models.utils import async_batching
from sl_ai_models.basic_model_interfaces.request_limited_model import RequestLimitedModel
from sl_ai_models.basic_model_interfaces.token_limited_model import TokenLimitedModel
from sl_ai_models.basic_model_interfaces.tokens_are_calculatable import TokensAreCalculatable


REQUEST_LIMITED_ERROR_MESSAGE = "Model must be RequestLimited"

######################################## CREATE TESTING INSTANCES ########################################

class NumModelMode(Enum):
    CALLS_FROM_ONE_MODEL = "calls_from_one_model"
    CALLS_FROM_MULTIPLE_MODELS = "calls_from_multiple_models"


class BurstMode(Enum):
    TESTING_ONLY_BURST = "testing_only_burst"
    TESTING_BURST_AND_AFTER_IS_CORRECT_RATE = "testing_burst_and_after_rate"
    TESTING_BURST_AND_AFTER_TAKES_TIME = "testing_burst_and_after_rate_takes_time"


class RunSetUp:
    def __init__(self, subclass: type[RequestLimitedModel],  mode: NumModelMode) -> None:
        self.subclass = subclass
        self.mode = mode

    @property
    def name_of_test(self) -> str:
        return f"{self.subclass.__name__} {self.mode.value}"


class RequestLimitTestResults:
    def __init__(self, duration: float, number_of_invokes: int, number_of_direct_calls: int) -> None:
        self.duration = duration
        self.number_of_invokes = number_of_invokes
        self.number_of_direct_calls = number_of_direct_calls

    @property
    def invokes_per_second(self) -> float:
        return self.number_of_invokes / self.duration

    @property
    def direct_calls_per_second(self) -> float:
        return self.number_of_direct_calls / self.duration


def get_testing_instances() -> list[tuple[str,RunSetUp]]:
    models_to_test = ModelsToTest.REQUEST_LIMITED_LIST
    tests = []
    for model in models_to_test:
        for mode in [NumModelMode.CALLS_FROM_ONE_MODEL, NumModelMode.CALLS_FROM_MULTIPLE_MODELS]:
            run_setup = RunSetUp(model, mode)
            name_of_run = run_setup.name_of_test
            tests.append((name_of_run,run_setup)) # Name of run apppended so it shows up in VSCode tests runner in an understandable way
    return tests



################################ TESTS ################################


@pytest.mark.parametrize(
    ("_", "run_setup"),
    get_testing_instances()
)
def test_burst_happens_quickly(mocker: Mock, _: str, run_setup: RunSetUp) -> None:
    number_of_calls = get_number_of_coroutines_for_initial_burst(run_setup.subclass)
    run_request_limit_test(mocker, run_setup, number_of_calls, BurstMode.TESTING_ONLY_BURST)


@pytest.mark.parametrize(
    ("_", "run_setup"),
    get_testing_instances()
)
def test_calls_after_burst_take_time(mocker: Mock, _: str, run_setup: RunSetUp) -> None:
    burst_calls = get_number_of_coroutines_for_initial_burst(run_setup.subclass)
    regular_calls = get_number_of_calls_that_takes_10s_to_run_given_model_rate(run_setup.subclass)
    number_of_calls = burst_calls + regular_calls
    run_request_limit_test(mocker, run_setup, number_of_calls, BurstMode.TESTING_BURST_AND_AFTER_TAKES_TIME)


############################################## Helper Functions ##############################################
def run_request_limit_test(mocker: Mock, run_setup: RunSetUp, number_of_calls_to_make: int, burst_mode: BurstMode) -> None:
    coroutines_to_test = setup_coroutines(run_setup, number_of_calls_to_make)
    mocked_direct_call = set_up_mocking_for_request_limit_tests(mocker, run_setup.subclass)

    if burst_mode == BurstMode.TESTING_BURST_AND_AFTER_TAKES_TIME:
        run_test_for__burst_and_after_takes_time(coroutines_to_test)
    else:
        run_other_burst_modes(run_setup, coroutines_to_test, burst_mode, mocked_direct_call)


def setup_coroutines(run_setup: RunSetUp, number_of_calls_to_make: int) -> list[Coroutine]:
    assert issubclass(run_setup.subclass, RequestLimitedModel), REQUEST_LIMITED_ERROR_MESSAGE
    AiModelMockManager.reinitialize_limiters(run_setup.subclass)
    coroutines_to_test = create_invoke_coroutines_for_test(run_setup, number_of_calls_to_make)
    return coroutines_to_test


def run_test_for__burst_and_after_takes_time(coroutines_to_test: list[Coroutine]) -> None:
    make_sure_to_take_longer_than_seconds = 5
    timed_coroutines = async_batching.wrap_coroutines_with_timeout(coroutines_to_test, make_sure_to_take_longer_than_seconds)
    if len(timed_coroutines) > 4000:
        pytest.skip("Skipping test because with our current usage we won't realistically hit this usage amount, and over this many coroutines messes up timeout because logging files become a resource bottleneck")

    logger.info(f"Running {len(timed_coroutines)} test coroutines")
    start_time = time.time()
    with pytest.raises(asyncio.TimeoutError):
        async_batching.run_coroutines(timed_coroutines)
    end_time = time.time()
    duration = end_time - start_time
    assert duration >= make_sure_to_take_longer_than_seconds, f"Didn't wait long enough. Duration: {duration}, Timeout: {make_sure_to_take_longer_than_seconds}"
    assert duration < make_sure_to_take_longer_than_seconds + 2, f"Waited too long. Duration: {duration}, Timeout: {make_sure_to_take_longer_than_seconds}"


def run_other_burst_modes(run_setup: RunSetUp, coroutines_to_test: list[Coroutine], burst_mode: BurstMode, mocked_direct_call: Mock) -> None:
    test_results = get_request_limit_test_results(run_setup.subclass, coroutines_to_test, mocked_direct_call)
    if burst_mode == BurstMode.TESTING_ONLY_BURST:
        logger.info(f"Got duration {test_results.duration}, with {test_results.number_of_direct_calls} direct calls over {test_results.number_of_invokes} coroutines")
        coroutine_testing.assert_resource_burst_is_short(run_setup.subclass.REQUESTS_PER_PERIOD_LIMIT, test_results.duration)
    elif burst_mode == BurstMode.TESTING_BURST_AND_AFTER_IS_CORRECT_RATE:
        assert_rate_limit_not_exceeded_or_missed_for_model(run_setup.subclass, test_results)


def run_burst_for_model_subclass(run_setup: RunSetUp) -> None:
    number_of_calls_to_make = get_number_of_coroutines_for_initial_burst(run_setup.subclass)
    coroutines_to_test = create_invoke_coroutines_for_test(run_setup, number_of_calls_to_make)
    async_batching.run_coroutines(coroutines_to_test)


def set_up_mocking_for_request_limit_tests(mocker: Mock, subclass: type[RequestLimitedModel]) -> Mock:
    mocked_direct_call = AiModelMockManager.mock_ai_model_direct_call_with_predefined_mock_value(mocker, subclass)

    if issubclass(subclass, TokensAreCalculatable):
        AiModelMockManager.mock_input_to_tokens_with_value(mocker, subclass)

    return mocked_direct_call


def get_number_of_calls_that_takes_10s_to_run_given_model_rate(subclass: type[RequestLimitedModel]) -> int:
    time_tests_should_take_in_seconds = 10
    allowed_calls_per_second = subclass.REQUESTS_PER_PERIOD_LIMIT / subclass.REQUEST_PERIOD_IN_SECONDS
    number_of_calls_to_make = int(time_tests_should_take_in_seconds * allowed_calls_per_second)
    return number_of_calls_to_make


def get_number_of_coroutines_for_initial_burst(subclass: type[RequestLimitedModel]) -> int:
    return subclass.REQUESTS_PER_PERIOD_LIMIT


def create_invoke_coroutines_for_test(run_setup: RunSetUp, number_of_coroutines: int) -> list[Coroutine]:
    if run_setup.mode == NumModelMode.CALLS_FROM_MULTIPLE_MODELS:
        models_to_create = 2
    else:
        models_to_create = 1

    models = [run_setup.subclass() for _ in range(models_to_create)]

    coroutines = []
    for model in models:
        number_of_calls_for_model = number_of_coroutines // len(models)
        cheap_input = model._get_cheap_input_for_invoke()
        coroutines.extend(model.invoke(cheap_input) for _ in range(number_of_calls_for_model))

    return coroutines


def get_request_limit_test_results(subclass: type[RequestLimitedModel], coroutines_to_test: list[Coroutine], mocked_direct_call: Mock) -> RequestLimitTestResults:
    if issubclass(subclass, TokenLimitedModel):
        subclass._make_token_limiter_have_large_rate()
    number_of_calls_before_this_test = mocked_direct_call.call_count
    stats = coroutine_testing.find_stats_of_coroutine_run(coroutines_to_test)
    duration = stats.duration_in_seconds
    call_count_of_direct_call_mock = mocked_direct_call.call_count - number_of_calls_before_this_test
    number_of_invokes = len(coroutines_to_test)
    return RequestLimitTestResults(duration, number_of_invokes, call_count_of_direct_call_mock)


def assert_rate_limit_not_exceeded_or_missed_for_model(subclass: type[RequestLimitedModel], test_results: RequestLimitTestResults) -> None:
    allowed_calls_per_second = subclass.REQUESTS_PER_PERIOD_LIMIT / subclass.REQUEST_PERIOD_IN_SECONDS
    allowed_upper_bound = allowed_calls_per_second * 1.50
    allowed_lower_bound = allowed_calls_per_second * 0.9

    results_message = f"Expected {round(allowed_calls_per_second,2)} direct calls per second, got {round(test_results.direct_calls_per_second,2)} direct calls per second, and {round(test_results.invokes_per_second,4)} invokes per second over {round(test_results.duration,2)} seconds, and {test_results.number_of_direct_calls} direct calls, over {test_results.number_of_invokes} coroutines"
    logger.info(results_message)
    assert test_results.direct_calls_per_second < allowed_upper_bound, f"The calls per second was too high. {results_message}"
    assert test_results.direct_calls_per_second > allowed_lower_bound, f"The calls per second was too low. {results_message}"


from typing import Any, get_args, get_origin
import logging
logger = logging.getLogger(__name__)
from abc import ABC
from sl_ai_models.basic_model_interfaces.ai_model import AiModel
import json
from sl_ai_models.utils.ai_misc import try_function_till_tries_run_out, validate_complex_type, strip_code_block_markdown
from typing import TypeVar
from pydantic import BaseModel
import ast

T = TypeVar('T')


class OutputsText(AiModel, ABC):

    async def invoke_and_return_verified_type(self, input: Any, normal_complex_or_pydantic_type: type[T], allowed_invoke_tries_for_failed_output: int = 3) -> T:
        """
        Input should ask for the type of resulting object you want with no other words around it
        Retries if an invalid format is given

        ## Handles
        - normal types (e.g. str, int, list, dict, bool, union)
        - complex types (e.g. list[str], dict[str, int], list[tuple[dict,int]], etc.)
        - Pydantic model or list of models (e.g. MyPydanticModel, list[MyPydanticModel], but not list[tuple[MyPydanticModel,int]])

        ## Example
        Prompt: Give me a list of recipe ideas as a list of string

        Type asked for: list[str]

        AI Returns:


            [
            "recipe idea 1",
            "recipe idea 2"
            ]


        Function Returns: list[str] containing ["recipe idea 1", "recipe idea 2"]
        """
        return await try_function_till_tries_run_out(allowed_invoke_tries_for_failed_output, self.__invoke_and_transform_to_type, input, normal_complex_or_pydantic_type)



    async def invoke_and_unsafely_run_and_return_generated_code(self, input: Any, expected_output_type: type[T], allowed_invoke_tries_for_failed_output: int) -> tuple[T,str]:
        """
        Input should ask for a code block with no other words around it
        The code block should assign the return value to a variable called 'final_result'
        Returns the code output as a string and the final result as the expected output type
        """
        return await try_function_till_tries_run_out(allowed_invoke_tries_for_failed_output, self.__invoke_and_unsafely_run_generated_code, input, expected_output_type)


    async def __invoke_and_transform_to_type(self, input: Any, normal_complex_or_pydantic_type: type[T]) -> T:
        response: str = await self.invoke(input)
        cleaned_response = strip_code_block_markdown(response.strip())
        try:
            transformed_response = self.transform_response_to_type(cleaned_response, normal_complex_or_pydantic_type)
        except Exception as e:
            raise ValueError(f"Error transorming response to type {normal_complex_or_pydantic_type}: {e}. Response was: {cleaned_response}")
        if not validate_complex_type(transformed_response, normal_complex_or_pydantic_type):
            raise TypeError(f"Model did not return {normal_complex_or_pydantic_type}. Output was: {cleaned_response}")
        return transformed_response


    @classmethod
    def transform_response_to_type(cls, response: str, normal_complex_or_pydantic_type: type[T]) -> T:
        outer_type = get_origin(normal_complex_or_pydantic_type)
        inner_types = get_args(normal_complex_or_pydantic_type)

        try:
            is_list = (outer_type == list)
            is_list_of_pydantic_models = is_list and issubclass(inner_types[0], BaseModel)
        except TypeError:
            is_list_of_pydantic_models = False

        try:
            is_pydantic_model = issubclass(normal_complex_or_pydantic_type, BaseModel)
        except TypeError:
            is_pydantic_model = False

        if is_list_of_pydantic_models:
            pydantic_model_type: type[BaseModel] = inner_types[0]
            list_of_validated_objects: list[BaseModel] = []
            list_of_dicts_from_response = json.loads(response)
            for item in list_of_dicts_from_response:
                pydantic_object = pydantic_model_type.model_validate(item)
                list_of_validated_objects.append(pydantic_object)
            final_response = list_of_validated_objects
        elif is_pydantic_model:
            assert issubclass(normal_complex_or_pydantic_type, BaseModel)
            model_as_json = json.loads(response)
            response_as_model = normal_complex_or_pydantic_type.model_validate(model_as_json)
            final_response = response_as_model
        else:
            final_response = cls.turn_string_into_non_pydantic_python_data_structure(response)

        assert validate_complex_type(final_response, normal_complex_or_pydantic_type)
        return final_response # type: ignore <- validate_complex_type checks this, but mypy doesn't pick it up


    async def __invoke_and_unsafely_run_generated_code(self, input: Any, expected_output_type: type[T]) -> tuple[T,str]:
        model_response = await self.invoke(input)
        code = strip_code_block_markdown(model_response)

        global_vars = {
            '__builtins__': __builtins__  # Allow access to standard libraries
        }
        local_vars = {}

        try:
            exec(code, global_vars, local_vars)
        except Exception as e:
            raise RuntimeError(f"Error executing code: {e}. Code was {code}")

        evaluated_answer = local_vars.get('final_result')
        if evaluated_answer is None:
            raise ValueError(f"Code did not assign anything to final_result. Code was {code}")

        if not validate_complex_type(evaluated_answer, expected_output_type):
            raise TypeError(f"Code run did not return {expected_output_type}. Output was: {evaluated_answer}")

        return evaluated_answer, code


    @staticmethod
    def turn_string_into_non_pydantic_python_data_structure(response: str) -> Any:
        if response == "[]" or response == "" or response == "[\"\"]" or response == "['']":
            return []

        try:
            response_loaded_as_string = json.loads(response)
        except json.JSONDecodeError as e1:
            try:
                response_loaded_as_string = ast.literal_eval(response)
            except Exception as e2:
                raise ValueError(f"Model did not return a parsable list. Error1: {e1}, Error2: {e2}, response: {response}")

        return response_loaded_as_string


    @staticmethod
    def get_schema_format_instructions_for_pydantic_type(pydantic_type: type[BaseModel]) -> str:
        # Copy schema to avoid altering original Pydantic schema.
        schema = {k: v for k, v in pydantic_type.model_json_schema().items()}

        reduced_schema = schema
        if "title" in reduced_schema:
            del reduced_schema["title"]
        if "type" in reduced_schema:
            del reduced_schema["type"]
        schema_str = json.dumps(reduced_schema)

        return _PYDANTIC_FORMAT_INSTRUCTIONS.format(schema=schema_str)



_PYDANTIC_FORMAT_INSTRUCTIONS = """
The output should be formatted as a JSON instance that conforms to the JSON schema below.

As an example, for the schema {{"properties": {{"foo": {{"title": "Foo", "description": "a list of strings", "type": "array", "items": {{"type": "string"}}}}}}, "required": ["foo"]}}
the object {{"foo": ["bar", "baz"]}} is a well-formatted instance of the schema. The object {{"properties": {{"foo": ["bar", "baz"]}}}} is not well-formatted.

Here is the output schema:
```
{schema}
```

Make sure to return an instance of the JSON (no other words, preface, or conclusion). Do not return the schema itself.
"""


from tests.models_to_test import ModelsToTest
from tests.ai_mock_manager import AiModelMockManager
from sl_ai_models.basic_model_interfaces.outputs_text import OutputsText
from sl_ai_models.basic_model_interfaces.ai_model import AiModel
from sl_ai_models.gpt4o import Gpt4o
from sl_ai_models.utils.response_types import TextTokenCostResponse
from sl_ai_models.utils.ai_misc import validate_complex_type

from unittest.mock import Mock
import asyncio
from typing import Coroutine, Any
import pytest
from pydantic import BaseModel
import logging
logger = logging.getLogger(__name__)

OUTPUTS_TEXT_ERROR_MESSAGE = "Model must be OutputsText"

@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.OUTPUTS_TEXT
)
def test_errors_if_does_not_return_expected_values(mocker: Mock, subclass: type[AiModel]) -> None:
    if not issubclass(subclass, OutputsText):
        raise ValueError(OUTPUTS_TEXT_ERROR_MESSAGE)

    AiModelMockManager.mock_ai_model_direct_call_with_predefined_mock_value(mocker, subclass) # We assume the default mock value does not ask for only a number as an answer
    ai_model = subclass()
    cheap_input = ai_model._get_cheap_input_for_invoke()
    with pytest.raises(Exception):
        asyncio.run(ai_model.invoke_and_unsafely_run_and_return_generated_code(cheap_input,int,1))



class TestPydanticModel(BaseModel):
    int_value: int
    float_value: float
    list_value: list[int]


class SubPydanticModel2(BaseModel):
    str_value: str


class TestPydanticModel2(BaseModel):
    str_value: str
    sub_model: SubPydanticModel2

instance_of_test_model_1 = TestPydanticModel(int_value=1, float_value=1.0, list_value=[1, 2, 3])
instance_of_test_model_2 = TestPydanticModel2(str_value="hello", sub_model=SubPydanticModel2(str_value="hello"))




@pytest.mark.parametrize(
    ("type_to_return", "mock_value_for_invoke", "expected_output"),
    [
        (int, "1", 1),
        (float, "1.0", 1.0),
        (str, "\"string\"", "string"),
        (list, "[1,2,3]", [1,2,3]),
        (dict, "{\"key\":\"value\"}", {"key": "value"}),
        (bool, "True", True),
        (tuple[int,int], "(1,2)", (1,2)),
        (list[int], "[1,2,3]", [1,2,3]),
        (list[bool], "[True,False]", [True,False]),
        (list[str], "[\"string1\",\"string2\"]", ["string1","string2"]),
        (list[str], "[\"string1\",\n\"string2\"]", ["string1","string2"]),
        (list[str], "[\n\"string1\",\n\"string2\"\n]", ["string1","string2"]),
        (list[str], "[\"string1\",\"string2\",]", ["string1","string2"]),
        (list[str], "[\n\"string1\",\n\"string2\",\n]", ["string1","string2"]),
        (list[str], "['string1','string2']", ["string1","string2"]),
        (list[str], "['string1',\n'string2']", ["string1","string2"]),
        (list[str], "[\n'string1',\n'string2'\n]", ["string1","string2"]),
        (list[str], "['string1','string2',]", ["string1","string2"]),
        (list[str], "[\n'string1',\n'string2',\n]", ["string1","string2"]),
        (list[str], "[]", []),
        (list[str], "[\"\"]", []),
        (list[str], "['']", []),
        (list[int], "[]", []),
        (list[dict], "[{\"key\":\"value\"}]", [{"key": "value"}]),
        (TestPydanticModel, "{\"int_value\":1,\"float_value\":1.0,\"list_value\":[1,2,3]}", instance_of_test_model_1),
        (TestPydanticModel2, "{\"str_value\":\"hello\",\"sub_model\":{\"str_value\":\"hello\"}}", instance_of_test_model_2),
        (list[TestPydanticModel], "[{\"int_value\":1,\"float_value\":1.0,\"list_value\":[1,2,3]},{\"int_value\":1,\"float_value\":1.0,\"list_value\":[1,2,3]}]", [instance_of_test_model_1, instance_of_test_model_1]),
        (list[str], "```[\"string1\",\"string2\"]```" , ["string1","string2"]),
        (list[str], "```json\n[\"string1\",\"string2\"]```", ["string1","string2"]),
        (list[str], "```python[\"string1\",\"string2\"]```", ["string1","string2"]),
        (list[tuple[int,int]], "[(1,2),(3,4)]", [(1,2),(3,4)]),
        (list[list[list[str]]], "[[[\"string1\",\"string2\"]]]", [[["string1","string2"]]]),
        (list[dict[str,list[int]]], "[{\"key\":[1,2,3]},{\"key\":[1,2,3]}]", [{"key":[1,2,3]},{"key":[1,2,3]}]),
        (int | float, "1", 1),
        (int | float, "1.0", 1.0),
    ]
)
def test_type_verification_works_for_valid_types(mocker: Mock, type_to_return: type, mock_value_for_invoke: str, expected_output: Any) -> None:
    mock_the_value_output_of_invoke(mocker, Gpt4o, mock_value_for_invoke)
    ai_model = Gpt4o()
    cheap_input = ai_model._get_cheap_input_for_invoke()
    coroutine = ai_model.invoke_and_return_verified_type(cheap_input, type_to_return, allowed_invoke_tries_for_failed_output=1)
    assert_output_is_of_expected_type_and_value(coroutine, type_to_return, expected_output)


@pytest.mark.parametrize(
    ("type_to_return", "mock_value_for_invoke"),
    [
        (int, "1.0"),
        (float, "1"),
        (int | float, "Hello"),
        (dict, "[1,2,3]"),
        (list[int], "[1,2,3,\"hello\"]"),
        (str, "1"),
        (bool, "1"),
        (list[str], "[1,2,3]"),
        (TestPydanticModel, "{\"int_value\":\"Hello\",\"float_value\":1.0,\"list_value\":[1,2,3]}"),
    ]
)
def test_type_verification_fails_for_invalid_types(mocker: Mock, type_to_return: type, mock_value_for_invoke: str) -> None:
    mock_the_value_output_of_invoke(mocker, Gpt4o, mock_value_for_invoke)
    ai_model = Gpt4o()
    cheap_input = ai_model._get_cheap_input_for_invoke()
    coroutine = ai_model.invoke_and_return_verified_type(cheap_input, type_to_return, allowed_invoke_tries_for_failed_output=1)
    with pytest.raises(Exception):
        asyncio.run(coroutine)


@pytest.mark.parametrize(
    ("mock_value_for_invoke", "expected_output_type", "expected_value"),
    [
        ("```python\nfinal_result = {\"key\":\"value\"}\n```", dict, {"key": "value"}),
        ("```\nfinal_result = {\"key\":\"value\"}```", dict, {"key": "value"}),
        ("final_result = {\"key\":\"value\"}", dict, {"key": "value"}),
        ("a=1\nb=2\nfinal_result = a+b",int,3),
        ("import math\nfinal_result = math.sqrt(4)",float,2),
        ("print('hello')",Exception,None),
        ("raise Exception('error')",Exception,None),
        ("final_result = (1,2,3)",tuple[int,int,int],(1,2,3)),
        ("final_result = [1,2,3]",list[int], [1,2,3]),
    ]
)
def test_invoke_for_running_code(mocker: Mock, mock_value_for_invoke: str, expected_output_type: type, expected_value: Any) -> None:
    mock_the_value_output_of_invoke(mocker, Gpt4o, mock_value_for_invoke)
    ai_model = Gpt4o()
    cheap_input = ai_model._get_cheap_input_for_invoke()
    coroutine = ai_model.invoke_and_unsafely_run_and_return_generated_code(cheap_input, expected_output_type, 1)
    if expected_value is None:
        assert issubclass(expected_output_type, Exception)
        with pytest.raises(expected_output_type):
            asyncio.run(coroutine)
    else:
        result, code = asyncio.run(coroutine)
        assert validate_complex_type(result, expected_output_type)
        assert result == expected_value
        assert isinstance(code, str)



def test_schema_generation_works() -> None:
    class TestPydanticModel(BaseModel):
        int_value: int
        float_value: float
        list_value: list[int]

    format_instructions = Gpt4o().get_schema_format_instructions_for_pydantic_type(TestPydanticModel)
    logger.debug(format_instructions)
    assert "int_value" in format_instructions
    assert "float_value" in format_instructions
    assert "list_value" in format_instructions



def mock_the_value_output_of_invoke(mocker: Mock, ai_model: type[AiModel], mock_value: str) -> None:
    mock_response = TextTokenCostResponse(
        data=mock_value,
        total_tokens_used=1,
        prompt_tokens_used=1,
        completion_tokens_used=1,
        cost=1.0,
        model="mock model",
    )
    AiModelMockManager.mock_ai_model_direct_call_with_value(mocker, ai_model, mock_response)


def assert_output_is_of_expected_type_and_value(coroutine: Coroutine, expected_type: type, expected_value: Any) -> None:
    model_response = asyncio.run(coroutine)
    assert validate_complex_type(model_response, expected_type)
    assert model_response == expected_value



from __future__ import annotations
from abc import ABC
from sl_ai_models.basic_model_interfaces.ai_model import AiModel
import logging
logger = logging.getLogger(__name__)
import functools
from typing import Any, Callable, Coroutine, TypeVar
from sl_ai_models.resource_managers.refreshing_bucket_rate_limiter import RefreshingBucketRateLimiter
from sl_ai_models.basic_model_interfaces.tokens_are_calculatable import TokensAreCalculatable

T = TypeVar('T')

class TokenLimitedModel(AiModel,TokensAreCalculatable, ABC):
    TOKENS_PER_PERIOD_LIMIT: int = NotImplemented
    TOKEN_PERIOD_IN_SECONDS: int = NotImplemented
    _token_limiter: RefreshingBucketRateLimiter = NotImplemented


    def __init_subclass__(cls: type[TokenLimitedModel], **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        is_not_abstract = ABC not in cls.__bases__
        if is_not_abstract:
            if cls.TOKENS_PER_PERIOD_LIMIT is NotImplemented:
                raise NotImplementedError('You forgot to define TOKENS_PER_PERIOD_LIMIT')
            if cls.TOKEN_PERIOD_IN_SECONDS is NotImplemented:
                raise NotImplementedError('You forgot to define TOKEN_PERIOD_IN_SECONDS')
            cls._reinitialize_token_limiter()


    @classmethod
    def _reinitialize_token_limiter(cls) -> None:
        cls._token_limiter = NotImplemented
        cls._token_limiter = RefreshingBucketRateLimiter(cls.TOKENS_PER_PERIOD_LIMIT, cls.TOKENS_PER_PERIOD_LIMIT/cls.TOKEN_PERIOD_IN_SECONDS)


    @staticmethod
    def _wait_till_token_capacity_available(func: Callable[..., Coroutine[Any, Any, T]]) -> Callable[..., Coroutine[Any, Any, T]]:
        @functools.wraps(func)
        async def wrapper(self: TokenLimitedModel, *args, **kwargs) -> T:
            tokens_of_prompt = self.input_to_tokens(*args, **kwargs)
            await self._token_limiter.wait_till_able_to_acquire_resources(tokens_of_prompt)
            result = await func(self, *args, **kwargs)
            return result
        return wrapper


    @classmethod
    def _make_token_limiter_have_large_rate(cls) -> None:
        """
        NOTE: This method is only for testing purposes.
        """
        absurdly_large_capacity = 100000000
        cls._token_limiter = RefreshingBucketRateLimiter(absurdly_large_capacity,absurdly_large_capacity/1)



from __future__ import annotations
from abc import ABC
from sl_ai_models.basic_model_interfaces.ai_model import AiModel
from sl_ai_models.utils import async_batching
import logging
logger = logging.getLogger(__name__)
from typing import Any, Coroutine, TypeVar, Callable
import functools

T = TypeVar('T')


class TimeLimitedModel(AiModel, ABC):
    TIMEOUT_TIME: int = NotImplemented


    def __init_subclass__(cls: type[TimeLimitedModel], **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        if ABC not in cls.__bases__ and cls.TIMEOUT_TIME is NotImplemented:
            raise NotImplementedError('You forgot to define TIMEOUT_TIME')


    @staticmethod
    def _wrap_in_model_defined_timeout(func: Callable[..., Coroutine[Any, Any, T]]) -> Callable[..., Coroutine[Any, Any, T]]:
        @functools.wraps(func)
        async def wrapper(self: TimeLimitedModel, *args, **kwargs) -> T:
            async def wrapper2(self: TimeLimitedModel, *args, **kwargs) -> T:
                result = await func(self, *args, **kwargs)
                return result
            coroutine = wrapper2(self, *args, **kwargs)
            timed_coroutine =  async_batching.wrap_coroutines_with_timeout([coroutine], self.TIMEOUT_TIME)[0]
            return await timed_coroutine
        return wrapper





import logging
logger = logging.getLogger(__name__)
import tiktoken
from PIL import Image
from openai.types.chat import ChatCompletionMessageParam, ChatCompletionContentPartImageParam, ChatCompletionContentPartTextParam, ChatCompletionSystemMessageParam, ChatCompletionUserMessageParam
from openai.types.chat.chat_completion_content_part_image_param import ImageURL
from tiktoken import Encoding
from typing import Literal
from pydantic import BaseModel
import base64
import math
import re
from io import BytesIO
from urllib import request




class VisionMessageData(BaseModel):
    prompt: str
    b64_image: str
    image_resolution: Literal["auto", "low", "high"]

    def __str__(self) -> str:
        return f"Prompt: {self.prompt}, Resolution: {self.image_resolution}, Image: {self.b64_image[:10]}..."


SMALL_BASE64_IMAGE: str = "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAQAAAC1HAwCAAAAC0lEQVR42mNk+A8AAQUBAScY42YAAAAASUVORK5CYII="
CHEAP_VISION_DATA = VisionMessageData(
    prompt="Hi",
    b64_image=SMALL_BASE64_IMAGE,
    image_resolution="low"
)


class OpenAiUtils:

    @staticmethod
    def text_to_tokens_direct(text_to_tokenize: str, model: str) -> int:
        encoding = tiktoken.encoding_for_model(model)
        token_num = len(encoding.encode(text_to_tokenize))
        return token_num


    @staticmethod
    def messages_to_tokens(messages: list[ChatCompletionMessageParam], model: str) -> int:
        """
        This functions takes in a list of messages and returns the number of tokens in the messages for gpt-3.5-turbo
        Here is an OpenAI guide with the example code: https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
        See here for other types of tokenization: https://github.com/shobrook/openlimit/blob/master/openlimit/utilities/token_counters.py
        See here for gpt-vision code. This is an unpublished push to the examples: https://github.com/openai/openai-cookbook/pull/881/commits/555f5bb8b6d09d83fcc7a892562f97d7d1f085c7
        """
        try:
            encoding = tiktoken.encoding_for_model(model)
        except KeyError:
            logger.warning("model not found. Using cl100k_base encoding.")
            encoding = tiktoken.get_encoding("cl100k_base")

        num_tokens = 0
        for message in messages:
            num_tokens += OpenAiUtils.__message_to_tokens(message, model, encoding)
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens


    @staticmethod
    def __message_to_tokens(message: ChatCompletionMessageParam, model: str, encoding: Encoding) -> int:
        if OpenAiUtils.__determine_if_message_is_an_image_message(message):
            return OpenAiUtils.__turn_image_message_into_tokens(message, model, encoding)
        else:
            return OpenAiUtils.__turn_regular_message_into_tokens(message, model, encoding)


    @staticmethod
    def __determine_if_message_is_an_image_message(message: ChatCompletionMessageParam) -> bool:
        try:
            content = message["content"] # type: ignore
            content_types = [part["type"] for part in content] # type: ignore
            assert "image_url" in content_types
            return True
        except Exception:
            return False


    @staticmethod
    def __turn_regular_message_into_tokens(message: ChatCompletionMessageParam, model: str, encoding: Encoding) -> int:
        if model in {
            "gpt-3.5-turbo-0613",
            "gpt-3.5-turbo-16k-0613",
            "gpt-4-0314",
            "gpt-4-32k-0314",
            "gpt-4-0613",
            "gpt-4-32k-0613",
            }:
            tokens_per_message = 3
            tokens_per_name = 1
        elif model == "gpt-3.5-turbo-0301":
            tokens_per_message = 4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
            tokens_per_name = -1  # if there's a name, the role is omitted
        elif "gpt-3.5-turbo" in model:
            # gpt-3.5-turbo may update over time. Returning num tokens assuming gpt-3.5-turbo-0613.
            return OpenAiUtils.__turn_regular_message_into_tokens(message,"gpt-3.5-turbo-0613",encoding)
        elif "gpt-4" in model:
            # gpt-4 may update over time. Returning num tokens assuming gpt-4-0613
            return OpenAiUtils.__turn_regular_message_into_tokens(message, "gpt-4-0613", encoding)
        else:
            raise NotImplementedError(
                f"""num_tokens_from_messages() is not implemented for model {model}. See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tokens."""
            )
        num_tokens = tokens_per_message
        for key, value in message.items():
            if not isinstance(value, str):
                raise ValueError(f"Value of key {key} is not a string. The value is {value}")
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
        return num_tokens


    @staticmethod
    def __turn_image_message_into_tokens(message: ChatCompletionMessageParam, model: str, encoding: Encoding) -> int:
        if model not in ["gpt-4-vision-preview", "gpt-4o"]:
            raise NotImplementedError(f"num_tokens_from_messages() is not implemented for model {model}")

        num_tokens: int = 0
        for key, value in message.items():
            if isinstance(value, list):
                for item in value:
                    num_tokens += len(encoding.encode(item["type"]))
                    if item["type"] == "text":
                        num_tokens += len(encoding.encode(item["text"]))
                    elif item["type"] == "image_url":
                        num_tokens += OpenAiUtils.__calculate_tokens_of_image(item["image_url"]["url"], item["image_url"]["detail"])
            elif isinstance(value, str):
                num_tokens += len(encoding.encode(value))

        return num_tokens


    @staticmethod
    def __get_image_dimensions(image_url_or_b64: str) -> tuple[int,int]:
        # regex to check if image is a URL or base64 string
        url_regex = r'https?:\/\/(www\.)?[-a-zA-Z0-9@:%._\+~#=]{1,256}\.[a-zA-Z0-9()]{1,6}\b([-a-zA-Z0-9()@:%_\+.~#?&\/\/=]*)' #NOSONAR
        base_64_regex = r'data:image\/\w+;base64,'
        if re.match(url_regex, image_url_or_b64):
            response = request.urlopen(image_url_or_b64)
            image = Image.open(response)
            return image.size
        elif re.match(base_64_regex, image_url_or_b64):
            image_url_or_b64 = re.sub(r'data:image\/\w+;base64,', '', image_url_or_b64)
            image = Image.open(BytesIO(base64.b64decode(image_url_or_b64)))
            return image.size
        else:
            raise ValueError("Image must be a URL or base64 string.")


    @staticmethod
    def __calculate_tokens_of_image(image_url_or_b64: str, detail:str) -> int:
        # Constants
        LOW_DETAIL_COST = 85
        HIGH_DETAIL_COST_PER_TILE = 170
        ADDITIONAL_COST = 85

        if detail == 'low':
            # Low detail images have a fixed cost
            return LOW_DETAIL_COST
        elif detail == 'high':
            # Calculate token cost for high detail images
            width, height = OpenAiUtils.__get_image_dimensions(image_url_or_b64)
            # Check if resizing is needed to fit within a 2048 x 2048 square
            if max(width, height) > 2048:
                # Resize the image to fit within a 2048 x 2048 square
                ratio = 2048 / max(width, height)
                width = int(width * ratio)
                height = int(height * ratio)

            # Further scale down to 768px on the shortest side
            if min(width, height) > 768:
                ratio = 768 / min(width, height)
                width = int(width * ratio)
                height = int(height * ratio)

            # Calculate the number of 512px squares
            num_squares = math.ceil(width / 512) * math.ceil(height / 512)
            total_cost = num_squares * HIGH_DETAIL_COST_PER_TILE + ADDITIONAL_COST
            return total_cost
        else:
            # Invalid detail_option
            raise ValueError("Invalid detail_option. Use 'low' or 'high'.")


    @staticmethod
    def put_single_user_message_in_list_using_prompt(user_prompt: str) -> list[ChatCompletionMessageParam]:
        return [
            ChatCompletionUserMessageParam(
                role="user",
                content=user_prompt
            )
        ]


    @staticmethod
    def put_single_image_message_in_list_using_gpt_vision_input(vision_data: VisionMessageData) -> list[ChatCompletionMessageParam]:
        prompt: str = vision_data.prompt
        base64_image: str = vision_data.b64_image
        resolution: str = vision_data.image_resolution
        return [
            ChatCompletionUserMessageParam(
                role="user",
                content=[
                    ChatCompletionContentPartTextParam(
                        type="text",
                        text=prompt
                    ),
                    ChatCompletionContentPartImageParam(
                        type="image_url",
                        image_url=ImageURL(
                            url=f"data:image/png;base64,{base64_image}",
                            detail=resolution
                        )
                    )
                ]
            )
        ]


    @staticmethod
    def create_sytem_and_user_message_from_prompt(user_prompt: str, system_prompt: str) -> list[ChatCompletionMessageParam]:
        return [
            ChatCompletionSystemMessageParam(
                role="system",
                content=system_prompt
            ),
            ChatCompletionUserMessageParam(
                role = "user",
                content=user_prompt
            )
        ]


    @classmethod
    def create_system_and_image_message_from_prompt(cls, vision_message_data: VisionMessageData, system_prompt: str) -> list[ChatCompletionMessageParam]:
        image_message_as_list = cls.put_single_image_message_in_list_using_gpt_vision_input(vision_message_data)
        image_message = image_message_as_list[0]
        return [
            ChatCompletionSystemMessageParam(
                role="system",
                content=system_prompt
            ),
            image_message
        ]


from sl_ai_models.deepgram_nova2 import DeepgramNova2
from sl_ai_models.utils import file_manipulation
import asyncio
import logging
logger = logging.getLogger(__name__)


def test_deepgram_works_on_file_input() -> None:
    project_file_path = "tests/test_data/short-clip-buellers-life-moves-pretty-fast.wav"
    assert_deepgram_works_on_input(file_manipulation.get_absolute_path(project_file_path))


def test_deepgram_works_on_url() -> None:
    url = "https://static.deepgram.com/examples/Bueller-Life-moves-pretty-fast.wav"
    assert_deepgram_works_on_input(url)


def assert_deepgram_works_on_input(input: str) -> None:
    model = DeepgramNova2()
    response = asyncio.run(model.invoke(input))
    logger.debug(f"Response: {response}")
    assert response is not None, "Response is None"





from pydantic import BaseModel
from typing import Any


class ModelResponse(BaseModel):
    data: Any


class TextTokenResponse(ModelResponse):
    data: str
    prompt_tokens_used: int
    completion_tokens_used: int
    total_tokens_used: int
    model: str


class TextTokenCostResponse(TextTokenResponse):
    cost: float



from __future__ import annotations
from sl_ai_models.resource_managers.hard_limit_manager_with_predictions import PredictThenResolveHardLimitManager
from sl_ai_models.resource_managers.hard_limit_manager_with_predictions import UsagePrediction # For other files to easily import from this file #NOSONAR
from sl_ai_models.resource_managers.hard_limit_manager import HardLimitExceededError # For other files to easily import from this file #NOSONAR


class MonetaryCostManager(PredictThenResolveHardLimitManager):
    """
    This class is a subclass of HardLimitManager that is specifically for monetary costs.
    Assume every cost is in USD
    """

    def __enter__(self) -> MonetaryCostManager:
        super().__enter__()
        return self


from sl_ai_models.model_archetypes.perplexity_text_model import PerplexityTextModel
from typing import Final
from sl_ai_models.utils.ai_misc import clean_indents  # Keep this import here for easier imports into other files so prompts can keep proper indentation levels in code # NOSONAR
from typing import TypeVar
T = TypeVar("T")


class Perplexity(PerplexityTextModel):
    MODEL_NAME: Final[str] = "llama-3.1-sonar-huge-128k-online"
    REQUESTS_PER_PERIOD_LIMIT: Final[int] = 16 # Technically 20, but has problems at 20 for some reason
    REQUEST_PERIOD_IN_SECONDS: Final[int] = 60
    TIMEOUT_TIME: Final[int] = 120
    TOKENS_PER_PERIOD_LIMIT: Final[int] = 2000000
    TOKEN_PERIOD_IN_SECONDS: Final[int] = 60
    PRICE_PER_TOKEN: Final[float] = 0.000005
    PRICE_PER_REQUEST: Final[float] = 0.005




from sl_ai_models.basic_model_interfaces.ai_model import AiModel
from sl_ai_models.basic_model_interfaces.time_limited_model import TimeLimitedModel
from sl_ai_models.basic_model_interfaces.retryable_model import RetryableModel
import asyncio
import pytest
from unittest.mock import Mock
from tests.models_to_test import ModelsToTest
from tests.ai_mock_manager import AiModelMockManager
import logging
logger = logging.getLogger(__name__)
import time

TIME_LIMITED_ERROR_MESSAGE = "Model must be TimeLimited"


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.TIME_LIMITED_LIST
)
def test_ai_model_successfully_times_out(mocker: Mock, subclass: type[AiModel]) -> None:
    if not issubclass(subclass, TimeLimitedModel):
        raise ValueError(TIME_LIMITED_ERROR_MESSAGE)

    mock_timeout_time = 10
    original_timeout_time = subclass.TIMEOUT_TIME
    subclass.TIMEOUT_TIME = mock_timeout_time

    AiModelMockManager.mock_ai_model_direct_call_with_long_wait(mocker, subclass)
    if issubclass(subclass, RetryableModel):
        model = subclass(allowed_tries=1)
    else:
        model = subclass()
    model_input = model._get_cheap_input_for_invoke()

    try:
        start_time = time.time()
        with pytest.raises(asyncio.exceptions.TimeoutError):
            asyncio.run(model.invoke(model_input))
        end_time = time.time()
        duration = end_time - start_time
        assert duration >= mock_timeout_time, f"Didn't wait long enough. Duration: {duration}, Timeout: {mock_timeout_time}"
        assert duration < mock_timeout_time + 3, f"Waited too long. Duration: {duration}, Timeout: {mock_timeout_time}"
    finally:
        subclass.TIMEOUT_TIME = original_timeout_time


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.TIME_LIMITED_LIST
)
def test_ai_model_does_not_time_out_when_run_time_less_than_timeout_time(mocker: Mock, subclass: type[AiModel]) -> None:
    if not issubclass(subclass, TimeLimitedModel):
        raise ValueError(TIME_LIMITED_ERROR_MESSAGE)

    AiModelMockManager.mock_ai_model_direct_call_with_predefined_mock_value(mocker, subclass)
    model = subclass()
    min_timeout_time = 3
    if model.TIMEOUT_TIME < min_timeout_time:
        raise ValueError(f"TIMEOUT_TIME must be greater than {min_timeout_time} since the mock function still takes time")
    model_input = model._get_cheap_input_for_invoke()
    response = asyncio.run(model.invoke(model_input))
    assert response is not None



from __future__ import annotations
import logging
logger = logging.getLogger(__name__)
from sl_ai_models.utils.response_types import TextTokenCostResponse
from abc import ABC
from sl_ai_models.model_archetypes.openai_text_model import OpenAiTextToTextModel
from openai.types.chat import ChatCompletionMessageParam
import os
from openai import AsyncOpenAI
from langchain_community.chat_models.perplexity import ChatPerplexity
from langchain_core.messages.utils import convert_to_messages
from sl_ai_models.basic_model_interfaces.priced_per_request import PricedPerRequest

class PerplexityTextModel(OpenAiTextToTextModel, PricedPerRequest, ABC):
    PRICE_PER_TOKEN: float


    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self.api_key is None:
            perplexity_api_key = os.getenv("PERPLEXITY_API_KEY")
            assert perplexity_api_key is not None, "PERPLEXITY_API_KEY is not set"
            self.api_key = perplexity_api_key
        self._OPENAI_ASYNC_CLIENT = AsyncOpenAI(
            api_key=self.api_key,
            base_url="https://api.perplexity.ai",
            max_retries=0 # Retry is implemented locally
        )


    def __init_subclass__(cls: type[PerplexityTextModel], **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        if ABC not in cls.__bases__:
            if cls.PRICE_PER_TOKEN is NotImplemented:
                raise NotImplementedError('You forgot to define PRICE_PER_TOKEN')
            if cls.PRICE_PER_REQUEST is NotImplemented:
                raise NotImplementedError('You forgot to define PRICE_PER_REQUEST')


    def input_to_tokens(self, prompt: str) -> int:
        messages: list[ChatCompletionMessageParam] = self._turn_model_input_into_messages(prompt)
        chat = ChatPerplexity(client=self._OPENAI_ASYNC_CLIENT, api_key=self.api_key, timeout=self.TIMEOUT_TIME)
        langchain_messages = convert_to_messages(messages) #type: ignore
        tokens = chat.get_num_tokens_from_messages(langchain_messages)
        adjustment = -2 * len(messages) # Experimentally the actual tokens always seem to be 2 less more than the tokens caculated (it used to be 47 more)
        adjusted_tokens = tokens + adjustment
        if adjusted_tokens < 0:
            raise ValueError(f"adjusted_tokens is less than 0. tokens: {tokens}, adjustment: {adjustment}")
        return adjusted_tokens


    def caculate_cost_from_tokens(self, prompt_tkns: int, completion_tkns: int) -> float:
        """
        NOTE: Perplexity cost is not dependent on completion versus prompt differences
        NOTE: There is a Per-Request cost added to this function
        """
        total_tokens = prompt_tkns + completion_tkns
        cost = total_tokens * self.PRICE_PER_TOKEN + self.PRICE_PER_REQUEST
        return cost


    @classmethod
    def _get_mock_return_for_direct_call_to_model_using_cheap_input(cls) -> TextTokenCostResponse:
        data= 'The term "hi" has multiple meanings and uses:\n\n1. **Interjection**: It is commonly used as a greeting, similar to "hello" or "hey." For example, "Hi, how are you?".\n\n2. **Abbreviation**: It can stand for "Hawaii" or "Hawaiian Islands," especially in contexts like zip codes. In meteorology, it can also mean "heat index" or "humidity index".\n\n3. **Informal Adjective**: It is an informal, simplified spelling of "high," often used in terms like "hi-fi" (high fidelity).\n\n4. **Pop Culture**: "Hi" is also the title of a song by Hannah Diamond, a popstar and image-maker known for her work with PC Music.\n\n5. **Brand Name**: "Hi" is the name of a Web3 neobank that allows users to trade, save, and spend cryptocurrencies and fiat currencies. However, the UK operations of this platform have been ceased.\n\n6. **YouTube Content**: There are various YouTube videos titled "Hi" that address different topics, including personal messages from content creators and music videos.'
        prompt_tokens = 1
        completion_tokens = 254
        total_cost = cls().caculate_cost_from_tokens(prompt_tokens, completion_tokens)
        total_tokens = prompt_tokens + completion_tokens
        return TextTokenCostResponse (
            data=data,
            prompt_tokens_used=prompt_tokens,
            completion_tokens_used=completion_tokens,
            total_tokens_used=total_tokens,
            model=cls.MODEL_NAME,
            cost=total_cost
        )





from __future__ import annotations

from sl_ai_models.basic_model_interfaces.ai_model import AiModel
from sl_ai_models.resource_managers.refreshing_bucket_rate_limiter import RefreshingBucketRateLimiter

from abc import ABC
import functools
import logging
from typing import Any, Callable, Coroutine, TypeVar
logger = logging.getLogger(__name__)


T = TypeVar('T')


class RequestLimitedModel(AiModel, ABC):
    # For more thoughts on abstract class properties see https://stackoverflow.com/questions/45248243/most-pythonic-way-to-declare-an-abstract-class-property
    REQUESTS_PER_PERIOD_LIMIT: int = NotImplemented
    REQUEST_PERIOD_IN_SECONDS: int = NotImplemented
    _request_limiter: RefreshingBucketRateLimiter = NotImplemented


    def __init_subclass__(cls: type[RequestLimitedModel], **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        class_is_not_abstract = ABC not in cls.__bases__
        if class_is_not_abstract:
            if cls.REQUESTS_PER_PERIOD_LIMIT is NotImplemented:
                raise NotImplementedError('You forgot to define REQUESTS_PER_PERIOD_LIMIT')
            if cls.REQUEST_PERIOD_IN_SECONDS is NotImplemented:
                raise NotImplementedError('You forgot to define REQUEST_PERIOD_IN_SECONDS')
            cls._reinitialize_request_rate_limiter()


    @classmethod
    def _reinitialize_request_rate_limiter(cls) -> None:
        cls._request_limiter = NotImplemented
        cls._request_limiter = RefreshingBucketRateLimiter(cls.REQUESTS_PER_PERIOD_LIMIT, cls.REQUESTS_PER_PERIOD_LIMIT/cls.REQUEST_PERIOD_IN_SECONDS)


    @staticmethod
    def _wait_till_request_capacity_available(func: Callable[..., Coroutine[Any, Any, T]]) -> Callable[..., Coroutine[Any, Any, T]]:
        @functools.wraps(func)
        async def wrapper(self: RequestLimitedModel, *args, **kwargs) -> T:
            number_of_requests_being_made = 1
            await self._request_limiter.wait_till_able_to_acquire_resources(number_of_requests_being_made)
            result = await func(self, *args, **kwargs)
            return result
        return wrapper




import os
import json
import datetime as dat
import functools
from typing import Callable, Any

NAME_OF_PACKAGE: str = "sl_ai_models"


def get_absolute_path(path_in_package: str = "") -> str:
    """
    This function returns the absolute path of a file in the package
    If there is no parameter given, it will just give the absolute path of the package
    @param path_in_package: The path of the file in the package starting just after the package name (e.g. "data/claims.csv")
    """

    # Get the path of the current file
    current_file_path = os.path.abspath(__file__)

    # Get the absolute path of the package that the function is in
    package_path = os.path.dirname(current_file_path)
    while os.path.basename(package_path) != NAME_OF_PACKAGE:
        package_path = os.path.dirname(package_path)

    # Get the absolute path of the file in the package
    file_path = str(os.path.join(package_path, path_in_package))

    return file_path


def load_json_file(project_file_path: str) -> list[dict]:
    """
    This function loads a json file. Output can be dictionary or list of dictionaries (or other json objects)
    @param project_file_path: The path of the json file starting from top of package
    """
    full_file_path = get_absolute_path(project_file_path)
    with open(full_file_path, "r") as file:
        return json.load(file)


def load_text_file(file_path_in_package: str) -> str:
    full_file_path = get_absolute_path(file_path_in_package)
    with open(full_file_path, "r") as file:
        return file.read()



def write_json_file(file_path_in_package: str, input: list[dict]) -> None:
    json_string = json.dumps(input, indent=4)
    create_or_overwrite_file(file_path_in_package, json_string)


def create_or_overwrite_file(file_path_in_package: str, text: str) -> None:
    """
    This function writes text to a file, and creates the file if it does not exist
    """
    full_file_path = get_absolute_path(file_path_in_package)
    os.makedirs(os.path.dirname(full_file_path), exist_ok=True)
    with open(full_file_path, 'w') as file:
        file.write(text)


def create_or_append_to_file(file_path_in_package: str, text: str) -> None:
    """
    This function appends text to a file, and creates the file if it does not exist
    """
    full_file_path = get_absolute_path(file_path_in_package)
    os.makedirs(os.path.dirname(full_file_path), exist_ok=True)
    with open(full_file_path, 'a') as file:
        file.write(text)



def log_to_file(file_path_in_package: str, text: str, type: str = "DEBUG") -> None:
    """
    This function writes text to a file but adds a time stamp and a type statement
    """
    new_text = f"{type} - {dat.datetime.now()} - {text}"
    full_file_path = get_absolute_path(file_path_in_package)
    os.makedirs(os.path.dirname(full_file_path), exist_ok=True)
    with open(full_file_path, 'a+') as file:
        file.write(new_text + "\n")


def current_date_time_string() -> str:
    return dat.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')


if __name__ == "__main__":
    """
    This is the "main" code area, and can be used for quickly sandboxing and testing functions
    """
    pass


from sl_ai_models.basic_model_interfaces.ai_model import AiModel
from sl_ai_models.basic_model_interfaces.retryable_model import RetryableModel
import asyncio
import pytest
from unittest.mock import Mock
from tests.models_to_test import ModelsToTest
from tests.ai_mock_manager import AiModelMockManager
import inspect
import logging
logger = logging.getLogger(__name__)

RETRYABLE_ERROR_MESSAGE = "Model must be Retryable"


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.RETRYABLE_LIST
)
def test_ai_model_successfully_retries(mocker: Mock, subclass: type[AiModel]) -> None:
    if not issubclass(subclass, RetryableModel):
        raise ValueError(RETRYABLE_ERROR_MESSAGE)

    AiModelMockManager.mock_ai_model_direct_call_with_2_errors_then_success(mocker, subclass)
    number_of_retries = 3

    model = subclass()
    model.allowed_tries = number_of_retries
    model_input = model._get_cheap_input_for_invoke()
    response = asyncio.run(model.invoke(model_input))
    assert response is not None


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.RETRYABLE_LIST
)
def test_errors_when_runs_out_of_tries(mocker: Mock, subclass: type[AiModel]) -> None:
    if not issubclass(subclass, RetryableModel):
        raise ValueError(RETRYABLE_ERROR_MESSAGE)

    AiModelMockManager.mock_ai_model_direct_call_with_2_errors_then_success(mocker, subclass)
    model = subclass()
    model.allowed_tries = 2 # It should run out of retries with the first 2 errors
    model_input = model._get_cheap_input_for_invoke()
    with pytest.raises(Exception):
        asyncio.run(model.invoke(model_input))


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.RETRYABLE_LIST
)
def test_raises_error_on_tries_setter_if_invalid(subclass: type[AiModel]) -> None:
    if not issubclass(subclass, RetryableModel):
        raise ValueError(RETRYABLE_ERROR_MESSAGE)

    model = subclass()

    with pytest.raises(ValueError):
        model.allowed_tries = -1

    with pytest.raises(ValueError):
        model.allowed_tries = 0

    try:
        model.allowed_tries = 1
    except Exception:
        pytest.fail("Should not raise error on positive allowed_tries")


@pytest.mark.parametrize(
    "subclass",
    ModelsToTest.RETRYABLE_LIST
)
def test_raises_error_on_tries_init_if_invalid(subclass: type[AiModel]) -> None:
    if not issubclass(subclass, RetryableModel):
        raise ValueError(RETRYABLE_ERROR_MESSAGE)

    with pytest.raises(ValueError):
        subclass(allowed_tries=0)

    with pytest.raises(ValueError):
        subclass(allowed_tries=-1)

    try:
        subclass(allowed_tries=1)
    except Exception:
        pytest.fail("Should not raise error on positive allowed_tries")



import logging
import asyncio
from typing import Union, get_origin, get_args, Callable, Any
logger = logging.getLogger(__name__)


async def try_function_till_tries_run_out(tries: int, function: Callable, *args, **kwargs) -> Any:
    tries_left = tries
    while tries_left > 0:
        try:
            response = await function(*args, **kwargs)
            return response
        except Exception as e:
            tries_left -= 1
            if tries_left == 0:
                raise e
            logger.warning(f"Retrying function {function.__name__} due to error: {e}")
            await asyncio.sleep(1)


def retry_async_function(tries: int) -> Callable:
    def decorator(function: Callable) -> Callable:
        async def wrapper(*args, **kwargs) -> Any:
            return await try_function_till_tries_run_out(tries, function, *args, **kwargs)
        return wrapper
    return decorator


def validate_complex_type(value: Any, expected_type: Any) -> bool: #NOSONAR
    origin = get_origin(expected_type)
    args = get_args(expected_type)

    if origin is None:
        # Base case: expected_type is not a generic alias (like int, str, etc.)
        return isinstance(value, expected_type)

    if origin is Union:
        # Special handling for Union types (e.g., Union[int, str])
        return any(validate_complex_type(value, arg) for arg in args)

    if origin is tuple:
        # Special handling for tuple types
        if not isinstance(value, tuple) or len(value) != len(args):
            return False
        return all(validate_complex_type(v, t) for v, t in zip(value, args))

    if origin is list:
        # Special handling for list types
        if not isinstance(value, list):
            return False
        return all(validate_complex_type(v, args[0]) for v in value)

    if origin is dict:
        # Special handling for dict types
        if not isinstance(value, dict):
            return False
        key_type, value_type = args
        return all(validate_complex_type(k, key_type) and validate_complex_type(v, value_type) for k, v in value.items())

    # Fallback for other types
    return isinstance(value, expected_type)


def clean_indents(text: str) -> str:
    lines = text.split("\n")
    try:
        indent_level_of_first_line = find_indent_level_of_string(lines[0])
        indent_level_of_second_line = find_indent_level_of_string(lines[1])
        greatest_indent_level_of_first_two_lines = max(indent_level_of_first_line, indent_level_of_second_line)
    except IndexError:
        greatest_indent_level_of_first_two_lines = find_indent_level_of_string(lines[0])

    new_lines = []
    for line in lines:
        indent_level_of_line = find_indent_level_of_string(line)
        if indent_level_of_line >= greatest_indent_level_of_first_two_lines:
            new_line = line[greatest_indent_level_of_first_two_lines:]
        else:
            new_line = line.lstrip()
        new_lines.append(new_line)

    combined_new_lines = "\n".join(new_lines)
    return combined_new_lines


def find_indent_level_of_string(string: str) -> int:
    return len(string) - len(string.lstrip())


def strip_code_block_markdown(string: str) -> str:
    string = string.strip()
    if string.startswith("```json") and string.endswith("```"):
        string = string[7:-3].strip()
    elif string.startswith("```python") and string.endswith("```"):
        string = string[9:-3].strip()
    elif string.startswith("```") and string.endswith("```"):
        string = string[3:-3].strip()
    return string




from sl_ai_models.utils.openai_utils import OpenAiUtils
from sl_ai_models.model_archetypes.openai_text_model import OpenAiTextToTextModel
from abc import ABC
import logging
logger = logging.getLogger(__name__)
from sl_ai_models.utils.openai_utils import VisionMessageData, CHEAP_VISION_DATA
from sl_ai_models.utils.response_types import TextTokenCostResponse



class OpenAiVisionToTextModel(OpenAiTextToTextModel, ABC):

    async def invoke(self, input: VisionMessageData) -> str:
        response: TextTokenCostResponse =  await self._invoke_with_request_cost_time_and_token_limits_and_retry(input)
        return response.data


    async def _mockable_direct_call_to_model(self, input: VisionMessageData) -> TextTokenCostResponse:
        self._everything_special_to_call_before_direct_call()
        messages = self.create_messages_from_input(input)
        prompt_tokens: int = self.input_to_tokens(input)
        max_tokens: int = prompt_tokens + 1000
        response: TextTokenCostResponse = await self._call_online_model_using_api(messages, self.temperature, max_tokens)
        return response


    @staticmethod
    def _get_cheap_input_for_invoke() -> VisionMessageData:
        return CHEAP_VISION_DATA


    def create_messages_from_input(self, input: VisionMessageData) -> list:
        if self.system_prompt is None:
            return OpenAiUtils.put_single_image_message_in_list_using_gpt_vision_input(input)
        else:
            return OpenAiUtils.create_system_and_image_message_from_prompt(input, self.system_prompt)


    def input_to_tokens(self, vision_input: VisionMessageData) -> int:
        messages = self.create_messages_from_input(vision_input)
        tokens = OpenAiUtils.messages_to_tokens(messages, self.MODEL_NAME)
        return tokens


from sl_ai_models.basic_model_interfaces.ai_model import AiModel
from sl_ai_models.basic_model_interfaces.time_limited_model import TimeLimitedModel
from sl_ai_models.basic_model_interfaces.request_limited_model import RequestLimitedModel
from sl_ai_models.basic_model_interfaces.token_limited_model import TokenLimitedModel
from sl_ai_models.basic_model_interfaces.tokens_are_calculatable import TokensAreCalculatable
from unittest.mock import Mock
import asyncio
from sl_ai_models.utils.openai_utils import CHEAP_VISION_DATA, SMALL_BASE64_IMAGE
import logging
logger = logging.getLogger(__name__)
from typing import Any

class AiModelMockManager:
    SMALL_BASE64_IMAGE = SMALL_BASE64_IMAGE
    CHEAP_IMAGE_MESSAGE_DATA = CHEAP_VISION_DATA


    @staticmethod
    def get_direct_call_function_path_as_string(subclass: type[AiModel]) -> str:
        full_function_path = f'{subclass._mockable_direct_call_to_model.__module__}.{subclass._mockable_direct_call_to_model.__qualname__}'
        return full_function_path


    @staticmethod
    def mock_ai_model_direct_call_with_2_errors_then_success(mocker: Mock, subclass: type[AiModel]) -> Mock:
        mock_function = mocker.patch(AiModelMockManager.get_direct_call_function_path_as_string(subclass))
        mock_function.side_effect = [
            Exception("Mock Exception 1"),
            Exception("Mock Exception 2"),
            subclass._get_mock_return_for_direct_call_to_model_using_cheap_input()
        ]
        return mock_function


    @staticmethod
    def mock_ai_model_direct_call_with_long_wait(mocker: Mock, subclass: type[AiModel]) -> Mock:
        if not issubclass(subclass, TimeLimitedModel):
            raise ValueError("Model must be TimeLimited")

        timeout_time = subclass().TIMEOUT_TIME
        time_longer_than_timeout: int = timeout_time + 1

        def wait_longer_than_timeout(*args, **kwargs) -> Any:
            asyncio.run(asyncio.sleep(time_longer_than_timeout))
            return subclass._get_mock_return_for_direct_call_to_model_using_cheap_input()

        mock_function = mocker.patch(
            AiModelMockManager.get_direct_call_function_path_as_string(subclass),
            side_effect=wait_longer_than_timeout
        )
        return mock_function


    @staticmethod
    def mock_ai_model_direct_call_with_predefined_mock_value(mocker: Mock, subclass: type[AiModel]) -> Mock:
        input_value = subclass._get_mock_return_for_direct_call_to_model_using_cheap_input()
        mock_function = AiModelMockManager.mock_ai_model_direct_call_with_value(mocker, subclass, input_value)
        return mock_function


    @staticmethod
    def mock_ai_model_direct_call_with_value(mocker: Mock, subclass: type[AiModel], value: Any) -> Mock:
        direct_call_function = subclass._mockable_direct_call_to_model
        if isinstance(direct_call_function, Mock):
            return direct_call_function

        full_function_path = AiModelMockManager.get_direct_call_function_path_as_string(subclass)
        mock_function = mocker.patch(full_function_path)
        mock_function.return_value = value
        return mock_function


    @staticmethod
    def mock_ai_model_direct_call_with_only_errors(mocker: Mock, subclass: type[AiModel]) -> Mock:
        def throw_error(*args, **kwargs) -> None:
            raise RuntimeError("Mock Exception")

        mock_function = mocker.patch(
            AiModelMockManager.get_direct_call_function_path_as_string(subclass),
            side_effect=throw_error
        )
        return mock_function


    @staticmethod
    def mock_input_to_tokens_with_value(mocker: Mock, subclass: type[TokensAreCalculatable], mock_value: int = 1) -> Mock:
        input_to_tokens_function = subclass.input_to_tokens
        if isinstance(input_to_tokens_function, Mock):
            return input_to_tokens_function

        mock_function = mocker.patch(f"{subclass.input_to_tokens.__module__}.{subclass.input_to_tokens.__qualname__}")
        mock_function.return_value = mock_value
        return mock_function


    @staticmethod
    def mock_function_that_throws_error_if_test_limit_reached(mocker: Mock) -> Mock:
        mock_function = mocker.patch(f"{AiModel._increment_calls_then_error_if_testing_call_limit_reached.__module__}.{AiModel._increment_calls_then_error_if_testing_call_limit_reached.__qualname__}")
        return mock_function


    @staticmethod
    def mock_function_that_deactivates_langchain_tracing(mocker: Mock) -> Mock:
        mock_function = mocker.patch(f"{AiModel._deactivate_langchain_traces_if_in_testing_mode.__module__}.{AiModel._deactivate_langchain_traces_if_in_testing_mode.__qualname__}")
        return mock_function


    @staticmethod
    def reinitialize_limiters(subclass: type[RequestLimitedModel | TokenLimitedModel]) -> None:
        if issubclass(subclass, RequestLimitedModel):
            subclass._reinitialize_request_rate_limiter()

        if issubclass(subclass, TokenLimitedModel):
            subclass._reinitialize_token_limiter()






from sl_ai_models.utils.openai_utils import OpenAiUtils
from sl_ai_models.utils.response_types import TextTokenCostResponse
from sl_ai_models.model_archetypes.traditional_online_llm import TraditionalOnlineLlm

from openai import AsyncOpenAI
from openai.types.chat import ChatCompletionMessageParam
from langchain_community.callbacks.openai_info import get_openai_token_cost_for_model

from abc import ABC
import os
import logging

logger = logging.getLogger(__name__)


class OpenAiTextToTextModel(TraditionalOnlineLlm, ABC):

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._OPENAI_ASYNC_CLIENT = AsyncOpenAI(
            api_key=self.api_key,
            max_retries=0 # Retry is implemented locally
        )


    async def invoke(self, prompt: str) -> str:
        response: TextTokenCostResponse =  await self._invoke_with_request_cost_time_and_token_limits_and_retry(prompt)
        return response.data


    async def _mockable_direct_call_to_model(self, prompt: str) -> TextTokenCostResponse:
        self._everything_special_to_call_before_direct_call()
        messages = self._turn_model_input_into_messages(prompt)
        response: TextTokenCostResponse = await self._call_online_model_using_api(messages, self.temperature)
        return response


    def _turn_model_input_into_messages(self, prompt: str) -> list[ChatCompletionMessageParam]:
        if self.system_prompt is None:
            return OpenAiUtils.put_single_user_message_in_list_using_prompt(prompt)
        else:
            return OpenAiUtils.create_sytem_and_user_message_from_prompt(prompt, self.system_prompt)


    async def _call_online_model_using_api(self, messages: list[ChatCompletionMessageParam], temperature: float, max_tokens: int | None = None) -> TextTokenCostResponse:
        client = self._OPENAI_ASYNC_CLIENT

        response = await client.chat.completions.create(
            model=self.MODEL_NAME,
            messages=messages,
            temperature=temperature,
            max_tokens=max_tokens,
        )
        if response.choices[0].message.content is None:
            raise RuntimeError("The model failed to give an answer. response.choices[0].message.content is None")

        answer = response.choices[0].message.content
        usage_stats = response.usage
        if usage_stats is None:
            raise RuntimeError("usage_stats is None")
        prompt_tokens = usage_stats.prompt_tokens
        completion_tokens = usage_stats.completion_tokens
        total_tokens = usage_stats.total_tokens

        cost = self.caculate_cost_from_tokens(prompt_tkns=prompt_tokens, completion_tkns=completion_tokens)

        return TextTokenCostResponse(
            data=answer,
            prompt_tokens_used=prompt_tokens,
            completion_tokens_used=completion_tokens,
            total_tokens_used=total_tokens,
            model=self.MODEL_NAME,
            cost=cost
        )


    ################################## Methods For Mocking/Testing ##################################


    @classmethod
    def _get_mock_return_for_direct_call_to_model_using_cheap_input(cls) -> TextTokenCostResponse:
        cheap_input = cls._get_cheap_input_for_invoke()
        probable_output = "Hello! How can I assist you today?"

        model = cls()
        prompt_tokens = model.input_to_tokens(cheap_input)
        completion_tokens = OpenAiUtils.text_to_tokens_direct(probable_output, cls.MODEL_NAME)

        total_cost = model.caculate_cost_from_tokens(prompt_tkns=prompt_tokens, completion_tkns=completion_tokens)
        total_tokens = prompt_tokens + completion_tokens
        return TextTokenCostResponse (
            data=probable_output,
            prompt_tokens_used=prompt_tokens,
            completion_tokens_used=completion_tokens,
            total_tokens_used=total_tokens,
            model=cls.MODEL_NAME,
            cost=total_cost
        )


    @staticmethod
    def _get_cheap_input_for_invoke() -> str:
        return "Hi"



    ############################# Cost and Token Tracking Methods #############################

    def input_to_tokens(self, prompt: str) -> int:
        messages = self._turn_model_input_into_messages(prompt)
        tokens = OpenAiUtils.messages_to_tokens(messages, self.MODEL_NAME)
        return tokens


    def caculate_cost_from_tokens(self, prompt_tkns: int, completion_tkns: int) -> float:
        prompt_cost = get_openai_token_cost_for_model(self.MODEL_NAME, prompt_tkns, is_completion=False)
        completion_cost = get_openai_token_cost_for_model(self.MODEL_NAME, completion_tkns, is_completion=True)
        cost = prompt_cost + completion_cost
        return cost






from __future__ import annotations
from sl_ai_models.resource_managers.hard_limit_manager import HardLimitManager, HardLimitExceededError
from typing import Final
import logging
logger = logging.getLogger(__name__)

# TODO: Add prediction logging to the HardLimitManagerWithPredictions class (if log_usage_when_called is True)

class UsagePrediction:
    __id_counter: int = 0

    def __init__(self, amount: float) -> None:
        self.amount: Final[float] = amount
        UsagePrediction.__id_counter += 1
        self.id: Final[int] = UsagePrediction.__id_counter


class PredictThenResolveHardLimitManager(HardLimitManager):
    def __init__(self, hard_limit: float, log_usage_when_called: bool = False) -> None:
        super().__init__(hard_limit, log_usage_when_called)
        self.__predictions: list[UsagePrediction] = []


    def __enter__(self) -> PredictThenResolveHardLimitManager:
        super().__enter__()
        return self


    @property
    def actual_plus_predicted_usage(self) -> float:
        predicted_usage = sum(prediction.amount for prediction in self.__predictions)
        return self.current_usage + predicted_usage


    @classmethod
    def add_usage_prediction_to_parent_managers(cls, amount: float) -> UsagePrediction:
        if amount < 0:
            raise ValueError("Amount should be a positive number or zero")
        prediction = UsagePrediction(amount)
        for cost_manager in cls._active_limit_managers.get():
            if isinstance(cost_manager, PredictThenResolveHardLimitManager):
                cost_manager.add_usage_prediction_locally(prediction)
            else:
                raise TypeError("active_limit_managers has a manager of the wrong")
        return prediction


    def add_usage_prediction_locally(self, prediction: UsagePrediction) -> None:
        self.__predictions.append(prediction)


    @classmethod
    def raise_error_if_limit_would_be_reached(cls, amount_to_check_room_for: float = 0) -> None:
        super().raise_error_if_limit_would_be_reached(amount_to_check_room_for)
        if amount_to_check_room_for < 0:
            raise ValueError("Amount should be a positive number or zero")
        for cost_manager in cls._active_limit_managers.get():
            if isinstance(cost_manager, PredictThenResolveHardLimitManager):
                if cost_manager.hard_limit - cost_manager.actual_plus_predicted_usage < amount_to_check_room_for:
                    raise HardLimitExceededError(f"It is predictd that the usage amount {amount_to_check_room_for} would push usage beyond the limit. Predicted usage would be pushed to {cost_manager.actual_plus_predicted_usage + amount_to_check_room_for} exceeding the hard limit of {cost_manager.hard_limit}")
            else:
                raise TypeError("active_limit_managers has a manager of the wrong type")


    @classmethod
    def replace_predicted_usage_with_actual_usage_in_parent_managers(cls, prediction_to_replace: UsagePrediction, actual_usage: float) -> None:
        if actual_usage < 0:
            raise ValueError("Actual usage must be non-negative")
        for cost_manager in cls._active_limit_managers.get():
            if isinstance(cost_manager, PredictThenResolveHardLimitManager):
                cost_manager.replace_predicted_usage_with_actual_locally(prediction_to_replace, actual_usage)
            else:
                raise TypeError("active_limit_managers has a manager of the wrong type")


    def replace_predicted_usage_with_actual_locally(self, prediction_to_replace: UsagePrediction, actual_usage: float) -> None:
        if actual_usage < 0:
            raise ValueError("Actual usage must be non-negative")
        if prediction_to_replace not in self.__predictions:
            raise ValueError("Prediction not found")
        self.__predictions.remove(prediction_to_replace)
        self._current_usage += actual_usage
        if self.current_usage > self.hard_limit:
            logger.warning(f"Improperly predicted actual usage. The Hard limit is exceeded. Current Usage: {self.current_usage}. Hard Limit: {self.hard_limit}")
